# = CGIKit - CGI Application Framework
#
# == See also
# * www  - http://www.spice-of-life.net/download/cgikit/
# * mail - info@spice-of-life.net
#
# == License
# CGIKit is copyright (C) 2002- SPICE OF LIFE.
# It is a free software distributed under the GPL license.
#
# == Author
# SUZUKI Tetsuya <suzuki@spice-of-life.net>


# CKAdapter is an interface between CGIKit and a web server.
# This object has a reference to a request/response object
# created by a CKAdaptee object.
# 
# The role of CKAdapter is to provide the consistent interface to 
# CKApplication. The difference of the relatioship between 
# CGIKit and a web server is aboserbed by a CKAdaptee obejct. 
class CKAdapter
	attr_accessor :adaptee

        # interface corresponds to adaptee's class and filename.
	def initialize( interface )
		@interface = interface
		@adaptee   = _adaptee
	end

	def run( response )
		@adaptee.out response
	end

	def create_request
		@adaptee.request
	end

	def create_response
		@adaptee.response
	end

	private

	def _adaptee
		if    @interface == 'fastcgi'  then _interface = 'FastCGI'
		elsif @interface == 'mod_ruby' then _interface = 'ModRuby'
		else
			_interface = 'CGI'
		end

		klass = CKAdaptee.const_get( _interface )
		klass.new
	end
end
# The super class of CGIKit error classes.
class CKError < StandardError; end


# This class is one of the most important classes in CGIKit.
# CKApplication has two major roles. 
# 
# One role is that CKAplication provides parameters which 
# effect the whole behavior of CGIKit application.
# The other role is that CKApplication communicates with CKAdapter, 
# an interface between a web server and CGIKit.
# 
# When CGIKit receives a request from a client,
# CKApplication arranges its parameters and creates a CKAdapter object.
# Next, CKApplication gets a request object from the CKAdapter object. 
# In reponse to the request object, CKApplication creates 
# a response object and sends it to the CKAdapter object.  
#
# == Application Process
#
# === 1. Ready a request and response object
# First, CKApplication creates a CKAdapter object and gets a request ( CKRequest ) 
# object from it. Then, CKApplication creates a response ( CKResponse ) 
# object. 
#
# === 2. Event loop
# Next, CKApplication goes into event loop. In this loop, CKApplication creates 
# components depending on the request and invokes the specified method of the components
# until it returns an object whose class is not CKComponent.
#
# === 3. Return the response object to the CKAdapter object
# Finally, CKApplication converts the components created in event loop to HTML
# and adds the HTML to the CKResponse object created in 1.
# CKApplication sends the response object to the CKAdapter object.
# Then, the CKAdapter object displays the response object to a client.
# 
# == CKApplication Attributes
# 
# === Controlling application
# <b>target</b>::         Component to be displayed.
# <b>action</b>::         Invoked method name of the target component.
# <b>main</b>::           Main component. If target isn't specified,
#                         this component is shown.
#                         The default value is 'MainPage'.
# <b>path</b>::           Path of an application.
# <b>component_path</b>:: Path for components.
#                         Components are searched under this path.
# <b>document_root</b>::  Document root directory.
# <b>locale</b>::         Locale of an application in a transaction.
#                         If the locale is specified, CGIKit change
#                         a template for a component. The name of
#                         the template includes the component name and
#                         the locale name. Also, the template name is 
#                         seprated by underscore("_"). For example,
#                         if the locale is "ja" and
#                         the component name is "MainPage",
#                         the template name is "MainPage_ja.html".
# <b>master_locale</b>::  Main locale of an application in a transaction.
#                         If the master_locale is equal to the specified
#                         locale, the components use templates whose name
#                         doesn't include the locale name. For instance,
#                         if the master_locale is "ja", the locale is "ja"
#                         and the component name is "MainPage",
#                         the template name is "MainPage.html".
# <b>resource</b>::       Resource directory. If "file" attribute of
#                         CKImage is set, the CKImage object searches
#                         an image file from resource directory. ( classes
#                         which manages other resources will be developed. )
# <b>interface</b>::      Interface of adapter.
#                         The default value is an interface for CGI.
#
# === Debugging
# <b>debug</b>::            Shows the backtrace in an error page
#                           when true is set.
# <b>attributes_check</b>:: Checks required attributes of elements
#                           if true is set.
#                           If undefined attributes are used or
#                           required attributes don't exist,
#                           an error page is displayed when the value is true.
class CKApplication
	attr_accessor :target, :action, :main, :locale, :master_locale, \
		:baseurl, :document_root, :path, :component_path, :resource, \
		:adapter, :interface, :request, :response, :debug, :attributes_check

	VERSION          = 'CGIKit 1.0b5'
	PATH             = Dir.pwd
	COMPONENT_PATH   = Dir.pwd
	MAIN_PAGE        = 'MainPage'
	ATTRIBUTES_CHECK = true
	ADAPTER          = 'cgi'
	ERROR_PAGE       = 'CKErrorPage'

	class << self
		# Returns version of CGIKit.
		def version
			VERSION
		end
	end

	def initialize
		@path             = PATH
		@component_path   = COMPONENT_PATH
		@main             = MAIN_PAGE
		@attributes_check = ATTRIBUTES_CHECK
	end

	# Returns the name of the application without file extension.
	def name
		basename = File.basename( @path, '.*' )
	end

	def master_locale?
		if @locale == @master_locale then
			true
		elsif not @locale then
			true
		elsif ( ( not @master_locale ) and ( not @locale ) ) then
			true
		else
			false
		end
	end

	# This method runs the CGIKit application.
	#
	#  require 'cgikit'
	#
	#  app = CKApplication.new
	#  app.run
	def run
		begin
			_ready_request_response
			component = _event_loop
			@response.content = component.to_s
		rescue Exception => e
			@response = handle_error e
		end

		@adapter.run @response
	end

	def create_adapter
		unless self.interface then
			CKAdapter.new ADAPTER
		else
			CKAdapter.new self.interface
		end
	end

	def handle_error( error )
		error_page       = page ERROR_PAGE
		error_page.error = error
		error_page.debug = @debug
		error_page.create_response		
	end

	def page( name )
		CKElement.instance( name, self, nil, nil, nil )
	end

	private

	def _ready_request_response
	 	@adapter = create_adapter unless @adapter

		# Gets request and creates response.
		@request        = @adapter.create_request  unless @request
		@response       = @adapter.create_response unless @response
		@target         = @request['target']
		@action         = @request['action']
		@locale         = @request['locale']
		@master_locale  = @request['master_locale']
		@baseurl        = @request.script_name unless @baseurl
		@document_root  = @request.headers['DOCUMENT_ROOT']  \
			unless @document_root
		@target = @main.dup unless @target
	end

	def _event_loop
		result = page @target

		# Recreates a main component if the component includes CKPartsMaker.
		if ( result.is_a? CKPartsMaker ) and ( @action == nil )
			@target = result.substitute_page || @main
			result  = page @target
		end

		begin
			component = result
			result    = component.run
		end while result.is_a? CKComponent

		component
	end
end


# CKBinding.bind invokes a method ( action ) of a target ( object ).
# You can get the result of the method by CKBinding.bind like Kernel#send.
# The method call can accompany arguments.
#
# == Order of method search  is as follows
#
# === Getting methods
# 1. object.field
# 2. object.get_field
# 3. object.getField
# 4. object._field
# 5. object._get_field
# 6. object._getField
#
# === Setting methods
# 1. object.field=
# 2. object.set_field
# 3. object.setField
# 4. object._field=
# 5. object._set_field
# 6. object._setField
class CKBinding
	class << self
		def bind( target, action, *args )
			if not ( target and action ) then return nil end
			_object, _getOrder, _setOrder = _orders( target, action )

			# make decidion of accessors
			result = nil
			if   args.empty? then order = _getOrder
			else                  order = _setOrder end

			is_respond = false
			order.each { | method |
				if method =~ /\"(.*)\"/
					result = $1
				elsif _object.respond_to?( method ) == false
					next
				end

				if result then
					is_respond = true
					break
				elsif args.empty? then
					result     = _object.send method
					is_respond = true
					break
				else
					result     = _object.send( method, *args )
					is_respond = true
					break
				end
			}

			if is_respond == false
				raise NameError, \
					"No such variable or method - #{target.class}\##{action}"
			end
			result
		end

		private

		def _orders( target, action )
			_object = target

			if action.index( '.' ) then
				methods = action.split( '.' )
				message = methods.pop
				methods.each { | method |
					_object = _object.__send__( method )
				}
			else
				message = action
			end
		
			accessor = message[ 0, 1 ].upcase + message.slice( 1, action.size )
			_getOrder = [ message, "get_#{message}", "get#{accessor}", \
						  "_#{message}", "_get_#{message}", "_get#{accessor}" ]
			_setOrder = [ "#{message}=", "set#{accessor}", "set#{accessor}", \
						  "_#{message}=", "_set_#{message}", "_set#{accessor}" ]
			[ _object, _getOrder, _setOrder ]
		end
	end
end


# Creates URL to access components and methods.
class CKURL
	attr_accessor :secure, :target, :action, :oid, :query, :parent

	def initialize( application )
		@application = application
	end

	def url
		baseurl     = @application.baseurl
		locale      = @application.locale
		master      = @application.master_locale
		if target == nil then
			@target = @application.target
		end

		if ( secure and ( not baseurl =~ /^https:\/\//i ) ) then
			baseurl =~ /^http:\/\/(.*)/i
			if $1 then string = "https://#{$1}?"
			else       string = "https://#{baseurl}?" end
		else
			string = "#{baseurl}?"
		end

		if target       then string << "target=#{target}&" end
		if action       then string << "action=#{action}&" end
		if master       then string << "master_locale=#{master}&" end
		if locale       then string << "locale=#{locale}&" end
		if oid          then string << "oid=#{oid}&"       end
		if query        then string << query_string( query ) end
		string
	end

	def query_string( hash )
		str = ''
		hash.each { | key, value |
			str << "#{key}=#{CKUtilities.escape_url( value )}&"
		}
		str
	end
end

# The super class of dynamic element classes.
#
# Dynamic elements convert themselves to HTML.
# These are very important in development with CGIKit.
#
# CKElement returns HTML when to_s() is called.
# In components development, you bind elements to component's methods. 
# At runtime, the binding is executed by run(). 
# These two methods are expected to be overrided in the subclasses of CKElement.
#
# == About CKElement#init_element
# In initialize(), CKElement calls init_element().
# You can hook the inialization of elements by this method.
# The reason why element classes have init_element() is for convenience.
# It may be better to override init_element() than initialize().
#
# * CKElement#initialize
#  class MainPage < CKComponent
#    def initialize( app, parent, name, body )
#      super( app, parent, name, body )
#
#      # initializing code ...
#
#    end
#  end
#
# * CKElement#init_element
#
#  class MainPage < CKComponent
#    def init_element
#
#      # initializing code ...
#
#    end
#  end
#
# == Paths for searching elements and components
# CKElement objects, including components, are usually instantiated
# by CKElement.instance. This method loads elements/components and creates
# element objects. In this method, some paths are searched to require files.
#
# The searched paths:
# 1. CKApplicaton#component_path
# 2. ($LOAD_PATH)/cgikit/elements
# 3. ($LOAD_PATH)/cgikit/components
#
# The latter two paths are for extention elements or components.
# It is recommended that your own elements or components are installed in the 1st path.
#
# == Creation of custom elements
# If you create custom elements, the class of the custom elements must 
# inherit CKElement. Simple dynamic element can be created only 
# by overriding to_s(). to_s() is called when CGIKit
# converts the dynamic elements to HTML. If you add new attributes
# to your custom elements, run() must be overrided. 
class CKElement
	class UnknownElementError < CKError; end
	class AttributeError      < CKError; end

	attr_accessor :name, :body, :definition, :parent, :application

	class << self
		def instance( element, app, parent, name, body )
			body          = '' unless body
			component     = "#{app.component_path}/#{element}/#{element}"
			ext_element   = "cgikit/elements/#{element}"
			ext_component = "cgikit/components/#{element}/#{element}"
			paths         = [ ext_element, ext_component ]

			# starndard elements
			if Object.const_defined? element
				return Object.const_get(element).new(app,parent,name,body)
			# components
			elsif FileTest.exist? "#{component}.rb"
				require component
				obj = Object.const_get(element).new(app,parent,name,body)
				return obj
			end

			# check paths for extension components
			$LOAD_PATH.each { | load_path |
				paths.each { | path |
					ext_path = "#{load_path}/#{path}"
					if FileTest.exist?( "#{ext_path}.rb" )
						require path
						obj = Object.const_get(element).new(app,parent,name,body)
						obj.path = ext_path
						return obj
					else
						next
					end
				}
			}

			raise UnknownElementError, "No such '#{element}' element or component."
		end
	end

	def initialize( app, parent, name, body )
		self.application = app
		self.parent      = parent
		self.name        = name
		self.body        = body
		init_element
	end

	def init_element; end
	
	def definition
		parent.definitions[name]
	end

	def name_with_class
		"\"#{name}\":#{self.class}"
	end

	# Returns value of "name" attribute for form.
	def oid
		if self.class.to_s == 'CKSubmitButton' then
			"#{parent.class}.oid.#{name}"
		else
			"#{parent.class}.#{name}"
		end
	end
	
	# Fetches the result of the parent component's methods.
	# When is_binding is true, fetch() returns String without binding from . 
	def fetch( key, is_binding = true )
		data = definition[ key ]
		if is_binding == false
			data
		else
			_parse_ckd_value( parent, data )
		end
	end
	
	# This method will be overrided in subclasses.
	def run; end

	# Raises exception if an element has undefined attributes.
	def check_undefined_attributes( *attrs )
		attrs << 'oid'     unless attrs.include? 'oid'
		attrs << 'element' unless attrs.include? 'element'

		definition.each_key { | attr |
			if ( attrs.include? attr ) == false
				raise AttributeError, \
					"Attribute '#{attr}' is undefined attribute - #{name_with_class}"
			end
		}
	end

	# Raises exception if an element doesn't have required attributes.
	def check_required_attributes( *patterns )
		requires = 0
		attrs    = []

		patterns.each { | pattern |
			pattern.each { | attr |
				requires += 1 if definition.include? attr
			}

			return true if requires == pattern.size
			requires = 0
			attrs << ( "'" + pattern.join(', ') + "'" )
		}
		message =  "#{self.class} requires #{attrs.join(' or ')} "
		message << "of attribute - #{name_with_class}"
		raise AttributeError, message
	end
	
	private
	
	def _parse_ckd_value( target, string )
		if    string == 'true'         then value = true
		elsif string == 'false'        then value = false
		elsif string =~ /"(.*)"/       then value = $1
		elsif string =~ /'(.*)'/       then value = $1
		elsif ( ( not string =~ /[\D]/ ) and ( string =~ /[\d]/ ) ) then
			value = string.to_i
		else
			value = CKBinding.bind( target, string )
		end
		value
	end
end


# When you define your component, its class must inherit CKComponent in code
# file. The component dynamically renders HTML with template and binding file. 
#
# == Construction of component
# A component has 3 files in a directory whose name is the same as the
# component's name. These three files composite Model-View-Controller
# architecture.
#
# Template(View)::      A HTML file includes CGIKit tags ( <cgikit> ). One
#                       CGIKit tag corresponds to one dynamic element.
#
# Binding(Controller):: A definition file for elements is used by a component.
#                       The suffix, ckd, is abbreviation of "CGIKit Definition".
#                       The definition file connects Template to Code.
#
# Code(Model)::         A Ruby source where a component's class is defined. 
#                       It must inherit CKComponent. The name of the component's
#                       class must be the same as the component name. 
#
# ex) An application with a MainPage component
#  /cgi-bin/cgikit-app
#    cgikit-app.cgi
#    /MainPage
#      MainPage.html
#      MainPage.ckd
#      MainPage.rb
#
# In this case, CGIKit application is cgikit-app.cgi and one component,
# "MainPage". 
#
# === Template(View)
# Template is an ordinary HTML file, except CGIKit's own tag. 
# CGIKit tag has only one attribute, "name". Each <cgikit> tag in template 
# corresponds to one dynamic element.
#
# (MainPage.html)
#  <html>
#  <body>
#  <cgikit name="HelloWorld"/>
#  </body>
#  </html>
#
# === Binding(Controller)
# In a binding file, a type and attributes of the elements in the template
# are defined. Note that the format differs form Ruby.
#
# (MainPage.ckd)
#  HelloWorld : CKString {
#    value = sayHello;
#  }
# 
# ==== 1. ElementName : ElementType
# A name and type of the element must be separated with colon.
# In the example above, the element name is HelloWorld and 
# the element type is CKString. Note that HelloWorld is
# used in the template as "name" attribute of the CGIKit tag.
#
# ==== 2. {
# The blanket, "{", means beginning of the definition of the element attributes.
#
# ==== 3. attribute = value;
# In attributes defintion, you write some pairs of an element's attribute
# and its value. The value is method name or literal. 
# At the end of the pair, you must locate a semicolon.
# If you write string literal, you must enclose the string 
# with single or double quotation marks.
#
# ex)
# method::               <tt>attribute = method_name;</tt>
# string::               <tt>attribute = "string";</tt> or <tt>'string'</tt>
# figure(integer only):: <tt>attribute = 12345;</tt>
# true/false::           <tt>attribute = true;</tt>
#
# ==== 4. }
# Close the definition with a close-blanket.
#
# === Code(Model)
# A class of a component must be named the same name 
# as the component name, and inherit CKComponent.
# The class has the methods specified in the ckd file.
#
# (MainPage.rb)
#  class MainPage < CKComponent
#    def sayHello
#      "Hello World!"
#    end
#  end
#
# === Result
#  Hello World!
#
# In the HelloWorld example, the component is parsed in this order.
#
# 1. (Template) <cgikit name="HelloWorld"/>
# 2. (Binding) HelloWorld : CKString
# 3. (Binding) value = sayHello
# 4. (Code) MainPage#sayHello
# 5. (Code) "Hello World!"
# 6. (Browser) Hello World!
#
# == Calling methods of a component
# Some elements, for example, CKForm, CKSubmit and CKHyperlink,
# can call a method of a component when they receive some events.
# You can specify the called method by their "action" attribute.
#
# For example, if you bind Echo#echo method to a CKHyperlink object and
# the link created by the CKHyperlink object is clicked, Echo#echo is invoked.
# Please try running examples in samples/Examples by yourself.
#
# === Linking to other components
# Now, you can hook some events. Then, what can we do in the hooked event?
# When a submit button or hyperlink is clicked, a method bound to the 
# "action" attribute is called. If the result of the method is 
# a CKComponent object, the returned object is displayed as web page.
# If the returned object is not a CKComponent object, 
# the current component is displayed.
#
# CGIKit provides the easy way to create component objects.
# A method that creates an instance of component is page().
# It returns a object of the component specified by the name.
# You can link to other components by using this method.
#
# See example:
#
#  def other_page
#    # Create an instance of a component
#    other_page = page( 'OtherComponentName' )
#
#    # Control the object if needed
#    other_page.foo = true
#
#    # Return the component object
#    other_page
#  end
#
# This example displays a component, "OhterComponentName", 
# when a user clicks the element to which the method, "other_page", is bound.
# In this example, a OtherComponentName's object is created by page().
# After that, true is assigned to the "foo" attribute of the created object.
#
# Web page, the component object, can be controlled before returning to clients
# in this procedure.
#
# == Processing form data
# Normally, CGI scripts process form data by themselves.
# But, CGIKit applications don't have to do it.
#
# For example, when a web page is created from a template
# which has CKText, the "name" attribute of the input tag ( textarea ) is
# determined by the CKText. The value of the "name" attribute
# includes the ID. If a user clicks the submit button, the data of
# the input tag is assinged to the component by calling the
# method which the ID specifies. The method which the ID specifies
# is bound the CKText element to.
#
# <b>Note</b>: IDs are not unique. Its are created by joining names of
# components and elements that defined in the components.
# Don't use a defined element over and over. 
#
# == Nesting components
# You can use one component in another component like elements. It means that  
# it is possible to nest components. 
#
# === Parts maker
# You can write a component which converts itself to a
# part of a web page. In some cases, a web page is composed of these components.
# These components are called parts component. Parts componente
# is recommended to include CKPartsMaker. A component which includes
# CKPartsMaker isn't displayed even if its name is set to
# the CKApplication#target. See also CKPartsMaker section.
#
# === Example
# This is an example.  
#
# * Template of a nested component
#  <cgikit name="String"/>
#
# * Binding of a nested compoent
#  String : CKString {
#    value = string;
#  }
#
# * Code of a nested component
#  class NestedComponent < CKComponent
#    include CKPartsMaker
#    attr_accessor :string
#  end
#
# * Template of parent's component
#  <html>
#  ...
#  <cgikit name=Location>
#  ...
#  </html>
#
# * Binding of parent's component
#  Location : NestedComponent {
#    string = "This is the nested component.";
#  }
#
# * Result
#  <html>
#  ...
#  This is the nested component.
#  ...
#  </html>
#  
# As you can see, the nested component is bound to the method
# of the parent component.
class CKComponent < CKElement
	class FileNotFoundError < CKError; end

	attr_accessor :action, :path

	def request
		application.request
	end

	def response
		application.response
	end

	def create_response
		response = CKResponse.new nil
		response.content = self.to_s
		response
	end

	def html_string
		return @html_string if @html_string

		begin
			f = File.new( _view )
			@html_string = f.readlines.join
			f.close
		rescue
			raise FileNotFoundError, "Can't read a template file - #{_view}"
		end

		@html_string
	end
	
	def definitions
		return @definitions if @definitions

		unless FileTest.exist? _controller
			msg = "Can't read a binding file - #{_controller} #{self.class}"
			raise FileNotFoundError, msg
		end

		@definitions = CKDefinition.parse_ckd_file _controller
	end

	def initialize( app, parent, name, body )
		@path = "#{app.component_path}/#{self.class}/#{self.class}"
		super( app, parent, name, body )
	end

	# Returns an instance of the new component.
	def page( page )
		CKElement.instance( page, application, parent, nil, nil )
	end

	# Invokes the action after binding component's variables to attributes
	# for definitions, values from request to component's variables.
	def run
		_bind_attribute
		_bind_request
		if @action == nil then return end

		if self.class.to_s == application.target
			CKBinding.bind( self, @action )
		end
	end

	def to_s
		parser = CKHTMLParser.new( self, html_string )
		parser.parse
	end
	
	private

	def _view
		if application.master_locale? then
			"#{path}.html"
		else
			"#{path}_#{application.locale}.html"
		end
	end
	
	def _controller
		"#{path}.ckd"
	end
	
	def _bind_attribute
		if parent == nil then return end
		return unless parent
		return unless definition
		defaults = { "name" => true, "oid" => true, "element" => true }

		definition.each { | key, value |
			if defaults[key] then next end
			CKBinding.bind( self, key, _parse_ckd_value(parent,value) )
		}
	end
	
	def _bind_request
		request = application.request

		_target = self
		request.multiple_params.each { | key, value |
			( object, variable, method ) = key.split( '.' )
			
			if key == 'action' then
				@action = value[0];
				next
			# for CKRadioButton
			elsif ( ( variable == nil ) and ( method == nil ) and \
				( definitions[ value[0] ] ) ) then

				# checked or selection
				if definitions[ value[0] ][ 'checked' ] then
					_action = definitions[ value[0] ][ 'checked' ]
					_args   = true
				elsif definitions[ value[0] ][ 'selection' ] then
					_action = definitions[ value[0] ][ 'selection' ]
					__action = definitions[ value[0] ][ 'value' ]
					_args   = CKBinding.bind( self, __action )
				end

				CKBinding.bind( _target, _action, _args )
				next
			elsif object != self.class.to_s then
				next
			end
			
			# for CKSubmitButton
			if method then
				@action = definitions[method]['action']
				next
			end
			
			# for CKCheckbox
			if definitions[variable]['checked'] then
				setVariable = definitions[variable]['checked']
				value = true
			elsif definitions[variable]['selection'] then
				setVariable = definitions[variable]['selection']
			# for CKPopUpButton, CKBrowser
			elsif definitions[variable]['selected'] then
				setVariable = definitions[variable]['selected']
			else
				setVariable = definitions[variable]['value']
			end
			_action = setVariable
			
			# binding data
			if value == true then
			elsif definitions[ variable ][ 'element' ] == 'CKBrowser' then
				value.delete ""
			elsif value.size < 2 then
				value = value[0]
			end

			CKBinding.bind( _target, _action, value )
		}
	end
end


# CKContent is used in nested components.
# This element tag in the template is replaced
# with a part of the template of its grandparent component.
#
# == Example
# * Template of a nested component ( This component is the parent of CKContent element.)
#  <b><cgikit name=Content /></b>
#
# * Binding of a nested component
#  Content : CKContent {}
#
# * Code of a nested component
#  class NestedComponent < CKComponent
#    include CKPartsMaker
#  end
#
# * Template of a grandparent component
#  foo
#  <cgikit name=Nested>This is the nested component.</cgikit>
#  bar
#
# * Binding of a grandparent component
#  Nested : NestedComponent {}
#
# * Result
#  foo
#  <b>This is the nested component.</b>
#  bar
#
# Another example is available in samples/Examples.
# In samples/Examples, CKContent element is used in HeaderFooter, which 
# defines the header and footer for all of the appication's pages.
#
# <b>Note</b>: You can locate only one CKContent element in the
# template of a given component.
#
# == Bindings
# CKContent doesn't have any attribute.
class CKContent < CKElement
	def run
		if application.attributes_check
			check_undefined_attributes nil
		end
	end

	def to_s
		unless parent.parent
			return
		end

		parser = CKHTMLParser.new( parent.parent, parent.body )
		parser.parse
	end
end


# CKFrame generates frame tag in HTML.
#
# == Bindings
# Required attributes: <b>page</b>, <b>src</b> or <b>value</b>.
#
# The priorities of "src", "page" and "value" is "src", 
# "page", "value". 
# 
# <b>name</b>::  "name" attribute of HTML's <frame> tag.
# <b>page</b>::  Name of component that supplies the content
#                for the frame.
# <b>src</b>::   You specify the URL or file for the frame. 
# <b>value</b>:: Method that supplies the content. The parent of this element
#                must have the specified method.
class CKFrame < CKElement
	def run
		@attr_name  = fetch( 'name' )  # String
		@attr_page  = fetch( 'page' )  # String
		@attr_src   = fetch( 'src' )   # String
		@attr_value = fetch( 'value' ) # CKComponent

		if application.attributes_check
			check_undefined_attributes( 'name', 'page', 'src', 'value' )
			check_required_attributes( ['page'], ['src'], ['value'] )
		end
	end

	def to_s
		if @attr_src then
		elsif @attr_page then
			url = CKURL.new application
			url.target = @attr_page
			@attr_src = url.url
		elsif @attr_value then
			url = CKURL.new application
			url.target = parent.class
			url.action = @attr_value
			@attr_src = url.url
		end
		
		"<frame name=\"#@attr_name\" src=\"#@attr_src\">"
	end
end

# CKString shows the result of binding as string.
#
# == Bindings
# Required attributes: <b>value</b>
#
# <b>value</b>::  Text to be displayed.   
# <b>escape</b>:: Escapes HTML control characters in "value" if the "escape"
#                 is true. The default value is true.
# <b>empty</b>::  Text that is substituted for "value"
#                 when the "value" is nil.
class CKString < CKElement
	def run	
		@attr_value  = fetch( 'value' )  # String
		@attr_escape = fetch( 'escape' ) # true or false
		@attr_empty  = fetch( 'empty' )  # String
		@attr_escape = true unless @attr_escape

		if application.attributes_check
			check_undefined_attributes( 'value', 'escape', 'empty' )
			check_required_attributes( ['value'], ['empty'] )
		end
	end

	def to_s
		if ( @attr_empty ) and ( @attr_value == nil ) then
			@attr_value = @attr_empty
		end
		if ( @attr_escape ) and ( @attr_value ) then
			@attr_value = CKUtilities.escape_html @attr_value
		end

		@attr_value.to_s
	end
end


# CKHyperlink generates a hypertext link.
#
# == Bindings
# Required attributes: <b>action</b>, <b>href</b> or <b>page</b>.
#
# <b>action</b>::   Method to be invoked when the link is clicked.
# <b>disabled</b>:: Generates a non-active link if the value is true.
# <b>href</b>::     You specify the URL to other web page directly.
#                   This attribute prevails over "action" or "page" attribute.
# <b>page</b>::     Name of component to display when the link is clicked.
# <b>string</b>::   Text of the link. If the body of CKHyperlink tag is
#                   not empty, the body is displayed. For example,
#                   if the template includes "<cgikit name=link>foo</cgikit>",
#                   this element shows "foo" as the link.
# <b>target</b>::   target attribute of HTML's <a> tag. 
# <b>secure</b>::   Appends "https://" to the URL if the value is true.
#                   The default value is false.
# <b>query</b>::    Hash as the query string. The value of "query" attribute is
#                   converted to string. Then, the string is added to append
#                   the URL.
class CKHyperlink < CKElement
	def run
		@attr_action   = fetch( 'action', false ) # CKComponent
		@attr_disabled = fetch( 'disabled' )      # true or false
		@attr_href     = fetch( 'href' )          # String
		@attr_page     = fetch( 'page' )          # String
		@attr_string   = fetch( 'string' )        # String
		@attr_target   = fetch( 'target' )        # String
		@attr_secure   = fetch( 'secure' )        # true or false
		@attr_query    = fetch( 'query' )         # Hash

		if application.attributes_check
			check_undefined_attributes( 'action', 'disabled', 'href', \
				'page', 'string', 'target', 'secure', 'query' )
			check_required_attributes( ['action'], ['href'], ['page'] )
		end
	end

	def to_s
		# action
		url = CKURL.new application
		if @attr_page then target = @attr_page
		else               target = parent.class.to_s end
		url.target  = target
		url.parent  = parent
		url.action  = @attr_action
		url.secure  = @attr_secure
		url.query   = @attr_query if @attr_query
		@attr_action = url.url

		# to_s
		to_s = '<a'
		if   @attr_href    then to_s << " href=\"#@attr_href\""
		else                    to_s << " href=\"#@attr_action\""   end
		if   @attr_target  then to_s << " target=\"#@attr_target\"" end
		to_s << '>'
		if    @attr_string then to_s << @attr_string
		elsif body         then to_s << body end
		to_s << '</a>'
		
		if @attr_disabled == true then to_s = body end
		to_s
	end
end


# Creates an image tag.
#
# == Bindings
# Required attributes: <b>file</b> or <b>src</b>.
#
# <b>alt</b>::    Alternative text to the picture.
# <b>border</b>:: Size of image border.
# <b>file</b>::   Name of an image file searched from 
#                 resource directory(CKApplication#resource).
# <b>src</b>::    You specify an image file directly. This value prevails 
#                 over "file" attribute.
class CKImage < CKElement
	def run	
		@attr_alt    = fetch( 'alt' )    # String
		@attr_border = fetch( 'border' ) # Integer
		@attr_file   = fetch( 'file' )   # String
		@attr_src    = fetch( 'src' )    # String

		if application.attributes_check
			check_undefined_attributes( 'alt', 'border', 'file', 'src' )
			check_required_attributes( ['file'], ['src'] )
		end
	end

	def to_s
		if @attr_file then
			resource = application.resource || application.baseurl
			@attr_src = File.join( resource, @attr_file )
		end

		to_s = '<img'
		if @attr_src    then to_s << " src=\"#@attr_src\""       end
		if @attr_alt    then to_s << " alt=\"#@attr_alt\""       end
		if @attr_border then to_s << " border=\"#@attr_border\"" end
		to_s << '>'
	end
end


# Controls generating HTML.
#
# == Bindings
# Required attribute: <b>condition</b>
#
# <b>condition</b>:: If the value is true and "negate" is false, the body of the 
#                    CKCoditional tag is displayed.
# <b>negate</b>::    Inverts the meaning of the "condition".
#
# == Control table
#   condition  negate  result
#   true       false   show
#   false      false   not show
#   true       true    not show
#   false      false   show
class CKConditional < CKElement
	def run
		@attr_condition = fetch( 'condition' )
		@attr_negate    = fetch( 'negate' )

		if application.attributes_check
			check_undefined_attributes( 'condition', 'negate' )
			check_required_attributes( ['condition'] )
		end
	end

	def to_s
		to_s = ''
		if ( ( ( not @attr_negate    ) and @attr_condition ) or \
		     ( ( not @attr_condition ) and @attr_negate    ) ) then
			parser = CKHTMLParser.new( parent, body )
			to_s = parser.parse
		end
		to_s
	end
end


# Creates a fill-in form.
# Dynamic Elements of form, for example, CKBrowser, CKCheckbox, CKRadioButton,
# CKPopUpButton, CKText, CKTextField, CKSubmitButton and CKResetButton,
# are used within CKForm or HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>action</b>::  Method to invoke when a submit button in the form is clicked.
# <b>enctype</b>:: Encode type to send form data.
#                  You can use "POST", "GET" or "HEAD" as the value.
# <b>href</b>::    URL to which the browser directs.
# <b>target</b>::  Frame in a frameset that receive the page.
class CKForm < CKElement
	def run
		@attr_action  = fetch( 'action' )  # String
		@attr_enctype = fetch( 'enctype' ) # String
		@attr_href    = fetch( 'href' )    # String
		@attr_target  = fetch( 'target' )  # String
		if ( application.baseurl == nil ) and ( @attr_href == nil )
			@attr_href = application.request.script_name
		end
		if @attr_enctype == nil then @attr_enctype = 'POST' end
		unless ( @attr_enctype.upcase == 'POST' ) or \
			( @attr_enctype.upcase == 'GET' ) or \
			( @attr_enctype.upcase == 'HEAD' )
			@attr_enctype = 'POST'
		end

		if application.attributes_check
			check_undefined_attributes( 'action', 'method', 'enctype', \
				'href', 'target' )
		end
	end

	def to_s
		parser = CKHTMLParser.new( parent, body )
		value = parser.parse
		
		to_s = "<form method=\"#@attr_enctype\""
		if @attr_href    then to_s << " action=\"#@attr_href\""
		else                  to_s << " action=\"#{application.baseurl}\"" end
		if @attr_target  then to_s << " target=\"#@attr_target\"" end
		if @attr_enctype then to_s << " enctype=\"#@attr_enctype\"" end
		to_s << ">\n"
		to_s << "<input type=\"hidden\" name=\"target\" value=\"#{parent.class}\">\n"
		if @attr_action then
			to_s << "<input type=\"hidden\" name=\"action\" value=\"#@attr_action\">"
		end
		to_s << "#{value}\n</form>\n"
	end
end


# Creates a list whose multiple items can be selected.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>list</b>
#
# <b>escape</b>::    Escapes HTML control characters in "value"
#                    if the "escape" is true. The default value is true.
# <b>list</b>::      Array which is iterated through. The first object
# <b>default</b>::   The first item if no item is selected. 
# <b>selected</b>::  Items which are chosen from the list.
# <b>multiple</b>::  Multiple items of the list can be selected 
#                    if the value is true.
# <b>size</b>::      Size of item in appearance.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKBrowser < CKElement
	def run
		@attr_name     = oid
		@attr_escape   = fetch( 'escape' )
		@attr_list     = fetch( 'list' )
		@attr_default  = fetch( 'default' )
		@attr_selected = fetch( 'selected' )
		@attr_multiple = fetch( 'multiple' )
		@attr_size     = fetch( 'size' )
		@attr_disabled = fetch( 'disabled' )
		if @attr_selected == nil then @attr_selected = [] end

		if application.attributes_check
			check_undefined_attributes( 'escape', 'list', \
				'default', 'selected', 'multiple', 'size', 'disabled' )
			check_required_attributes( ['list'] )
		end
	end

	def to_s
		to_s = '<select'
		if @attr_name     then to_s << " name=\"#@attr_name\"" end
		if @attr_size     then to_s << " size=\"#@attr_size\"" end
		if @attr_multiple then to_s << " multiple" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">\n"
		if @attr_default then
			to_s << "<option value=\"\""
			if @attr_selected == nil then
				to_s << " selected=\"selected\""
			end
			to_s << ">#@attr_default</option>\n"
		end
		
		@attr_list.each { | item |
			if ( ( @attr_selected == item ) or ( @attr_selected.include?(item) ) ) then
				to_s << "<option selected=\"selected\">"
			else
				to_s << "<option>"
			end
			if @attr_escape == 'true' then
				item = CKMessage.escape( item )
			end
			to_s << "#{item}</option>\n"
		}
		to_s << '</select>'
	end
end


# Creates a checkbox.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: <b>selection</b> and <b>value</b>, or <b>checked</b>
#
# You use this element in two ways. One it the way to use "checked" attribute. 
# The other is the way to use both "selection" and "value" attributes. 
#
# <b>checked</b>::   If neither "value" nor "selection" attribute is nil and
#                    the value of "selection" is equal to that of "value",
#                    the check box is checked.
# <b>value</b>::     When the check box is checked,
#                    the value of "value" attribute is set to the component by
#                    the method specified by "selection" attribute.
# <b>selection</b>:: Object that the user chose from the check box.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKCheckbox < CKElement
	def run
		@attr_name      = oid
		@attr_checked   = fetch( 'checked' )   # true or false
		@attr_value     = fetch( 'value' )     # String
		@attr_selection = fetch( 'selection' ) # String
		@attr_disabled  = fetch( 'disabled' )  # true or false
		if ( ( @attr_value and @attr_selection ) and ( @attr_value == @attr_selection ) ) then
			@attr_checked = true
		end

		if application.attributes_check
			check_undefined_attributes( 'checked', 'value', 'selection', \
				'disabled' )
			check_required_attributes( ['selection','value'], ['checked'] )
		end
	end

	def to_s
		to_s =  "<input type=\"checkbox\" name=\"#@attr_name\""
		to_s << " value=\"#@attr_value\"" if @attr_value
		if @attr_checked  then to_s << " checked" end
		if @attr_disabled then to_s << " disabled" end
		to_s << '>'
	end
end


# Creates a pop-up menu.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>list</b>
#
# <b>escape</b>::    Escapes HTML control characters in the items of the list
#                    if the "escape" is true. The default value is true.
# <b>list</b>::      Array which is iterated through.
# <b>default</b>::   The first item if no item is selected.
# <b>selection</b>:: Item that are chosen from the selection list.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKPopUpButton < CKElement
	def run
		@attr_name     = oid
		@attr_escape   = fetch( 'escape' )      # true or false
		@attr_list     = fetch( 'list' )        # Enumerable
		@attr_default  = fetch( 'default' )     # String
		@attr_selected = fetch( 'selection' )   # String
		@attr_disabled = fetch( 'disabled' )    # true or false

		if application.attributes_check
			check_undefined_attributes( 'escape', 'list', \
				'default', 'selected', 'disabled' )
			check_required_attributes( ['list'] )
		end
	end

	def to_s
		to_s = '<select'
		if @attr_name     then to_s << " name=\"#@attr_name\"" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">\n"
		if @attr_default then
			to_s << "<option value=\"\""
			if @attr_selected == nil then
				to_s << " selected=\"selected\""
			end
			to_s << ">#@attr_default</option>\n"
		end
		
		@attr_list.each { | item |
			to_s << "<option"
			if @attr_selected == item then
				to_s << " selected=\"selected\""
			end
			
			# escape
			if @attr_escape then
				item = CKUtilities.escape_html( item )
			end
			
			to_s << ">#{item}</option>\n"
		}
		to_s << '</select>'
	end
end


# Creates a radio button.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: <b>selection</b> and <b>value</b>, or <b>checked</b>
#
# <b>checked</b>::   If neither "value" nor "selection" attribute is nil and
#                    the value of "selection" is equal to that of "value",
#                    the check box is checked.
# <b>value</b>::     When the check box is checked,
#                    the value of "value" attribute is set to the component by
#                    the method specified by "selection" attribute.
# <b>selection</b>:: Object that the user chose from the check box.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKRadioButton < CKElement
	def run
		@attr_name      = fetch( 'name' )      # String
		@attr_checked   = fetch( 'checked' )   # true or false
		@attr_value     = fetch( 'value' )     # String
		@attr_selection = fetch( 'selection' ) # String
		@attr_disabled  = fetch( 'disabled' )  # true or false
		if ( ( @attr_value and @attr_selection ) and \
			( @attr_value == @attr_selection ) ) then
			@attr_checked = true
		end

		if application.attributes_check
			check_undefined_attributes( 'checked', 'value', \
				'selection', 'disabled', 'name'  )
			check_required_attributes( ['selection','value'], ['checked'] )
		end
	end

	def to_s
		to_s = "<input type=\"radio\" name=\"#@attr_name\" value=\"#@attr_value\""
		if @attr_checked  then to_s << " checked" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">"
	end
end


# A CKRepeition object repeats its contents. 
#
# == Bindings
# Required attributes: <b>list</b> and <b>item</b>, or <b>count</b>
#
# <b>count</b>:: CKRepeition repeats its contents this number of times 
# <b>list</b>::  Array which is iterated through.
# <b>item</b>::  Current item when the list is iterated through.
# <b>index</b>:: Index of the current item.
class CKRepetition < CKElement
	def run
		@attr_count = fetch( 'count' )        # Integer
		@attr_item  = fetch( 'item', false )
		@attr_list  = fetch( 'list' )         # Enumerable
		@attr_index = fetch( 'index', false )
		if @attr_count == nil then @attr_count = @attr_list.size end

		if application.attributes_check
			check_undefined_attributes( 'count', 'item', 'list', 'index' )
			check_required_attributes( ['list', 'item'], ['count'] )
		end
	end

	def to_s
		to_s  = ''
		index = 0
		parser = CKHTMLParser.new( parent, body )

		@attr_list.each { | item |
			# set item
			CKBinding.bind( parent, @attr_item, item )
			
			# set index
			CKBinding.bind( parent, @attr_index, index )
			
			# parse
			to_s << parser.parse

			if index >= @attr_count then break end
			index += 1
		}
		to_s
	end
end


# Creates a reset button.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>value</b>:: Title of the button.
class CKResetButton < CKElement
	def run
		@attr_value  = fetch( 'value' ) # String
		unless @attr_value then @attr_value = 'Reset' end

		if application.attributes_check
			check_undefined_attributes nil
		end
	end

	def to_s
		"<input type=\"reset\" value=\"#@attr_value\">"
	end
end


# Creates a submit button.
# This element must be used within HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>action</b>::   Method to invoke when the button is clicked.
# <b>value</b>::    Title of the button.
# <b>disabled</b>:: If the value is true, the element appears but is not active.
#                   In addition, it doesn't send the form data 
#                   although the button is clicked.
class CKSubmitButton < CKElement
	def run
		@attr_name     = oid
		@attr_action   = fetch( 'action', false ) # String
		@attr_value    = fetch( 'value' )         # String
		@attr_disabled = fetch( 'disabled' )      # true/false
		if @attr_value == nil then @attr_value = 'Submit' end

		if application.attributes_check
			check_undefined_attributes( 'action', 'value', 'disabled' )
		end
	end

	def to_s
		to_s  = "<input type=\"submit\" name=\"#@attr_name\" value=\"#@attr_value\""
		if @attr_disabled then to_s << " disabled" end
		to_s << ">"
	end
end


# Creates a text area.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>value</b>
#
# <b>value</b>::   Value of the text area. 
# <b>columns</b>:: Column size.
# <b>rows</b>::    Row size.
class CKText < CKElement
	def run
		@attr_name    = oid
		@attr_value   = fetch( 'value' )   # String
		@attr_columns = fetch( 'columns' ) # Integer
		@attr_rows    = fetch( 'rows' )    # Integer

		if application.attributes_check
			check_undefined_attributes( 'value', 'columes', 'rows' )
			check_required_attributes( ['value'] )
		end
	end

	def to_s
		to_s = "<textarea name=\"#@attr_name\""
		if @attr_columns then to_s << " cols=\"#@attr_columns\"" end
		if @attr_rows    then to_s << " rows=\"#@attr_rows\""    end
		to_s << '>'
		if @attr_value   then to_s << @attr_value end
		to_s << '</textarea>'
	end
end


# Creates a text input field.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>value</b>
#
# <b>value</b>:: Value of the text field. If you set an accessor method to
#                this element, the form data is set to a component automatically
#                by the method.
# <b>type</b>::  Type of the text field. "text" is for a normal text input field,
#                "password" is for a password field and 
#                "hidden" is for a hidden field.
# <b>size</b>::  Size of the text field.
class CKTextField < CKElement
	def run
		@attr_name  = oid
		@attr_type  = fetch( 'type' )  # String
		@attr_value = fetch( 'value' ) # String
		@attr_size  = fetch( 'size' )  # Integer
		if @attr_type == nil then @attr_type = 'text' end

		if application.attributes_check
			check_undefined_attributes( 'type', 'value', 'size' )
			check_required_attributes( ['value'] )
		end
	end

	def to_s
		to_s = "<input type=\"#@attr_type\""
		if @attr_name  then to_s << " name=\"#@attr_name\"" end
		if @attr_value then to_s << " value=\"#@attr_value\"" end
		if @attr_size  then to_s << " size=\"#@attr_size\"" end
		to_s << '>'
	end
end


# CKPartsMaker is a Ruby module to deal with a component as parts of web page.
# A component which includes this module isn't displayed even if CGIKit recieves
# requests to show it as the target component ( CKApplication#target ).
#
# A name of parts component is recommended to have "Parts" at
# the end of the name to distinguish it from page component.
#
# == Usage
# Subclass of CKComponent includes CKPartsMaker.
#
#  class AnyParts < CKComponent
#    include CKPartsMaker
#    ...
#  end
#
# == Attributes
# <b>substitute_page</b>:: When CGIKit recieves requests to show component
#                          parts, CGIKit shows the page specified by
#                          this attribute. A main page of an application
#                          is displayed when the value is not defined.
module CKPartsMaker
	attr_accessor :substitute_page
end


# The super class of HTTP Request-Response classes.
class CKMessage
	attr_accessor :content, :headers, :http_version, :encoding
	attr_reader   :cookies

	EOL = "\r\n"

	def initialize( headers )
		@headers      = headers || {}
		@http_version = "1.1"
		@cookies      = CKUtilities.parse_cookie @headers['HTTP_COOKIE']
	end

	def add_cookie( cookie )
		@cookies << cookie
	end

	def remove_cookie( cookie )
		@cookies.delete cookie
	end
end


# CKRequest represents requests of HTTP headers.
#
# CKRequest returns form parameters and cookies.
# But, in many cases, this class isn't used directly to get parameters 
# because CGIKit sets these parameters to components 
# automatically in the process of instanciating components.
#
# == Getting a request object
# You can get a request object by CKApplication#request or CKComponent#request.
#
# == Parameters
# CKRequest has some methods to access parameters.
#
# <b>multiple_params[key]</b>::     Returns an array of parameters.
#                                   If key is missing, returns an empty array.
# <b>single_param(key), [key]</b>:: This method joins an array of parameters
#                                   into a string and returns it.
#                                   If key is missing, returns nil.
#
# == Cookies
# Cookies is an array which includes some CKCookie objects.
# See CKCookie section about details of cookie.
#
# <b>cookie(key)</b>::        Returns a CKCookie object whose key is the 
#                             same as the argument.
# <b>cookies</b>::            Returns an array of CKCookie objects.
# <b>cookie_value(key)</b>::  Returns the value of CKCookie object
#                             whose key is the the same as the argument.
# <b>cookie_values(key)</b>:: If the argument is nil(by default, argument is
#                             nil.), this method returns an array which has
#                             all the values of cookies. Otherwise, it returns
#                             an array which has the values of cookies
#                             specified by the argument.
#
# == HTTP Headers
# Some of HTTP headers are defined by instance methods.
# Call CKRequest#headers to get other CGI environment variables.
#
# <b>headers[key]</b>::      Returns a value of HTTP header.
# <b>accept</b>::            HTTP_ACCEPT
# <b>accept_charset</b>::    HTTP_ACCEPT_CHARSET 
# <b>accept_language</b>::   HTTP_ACCEPT_LANGUAGE
# <b>auth_type</b>::         AUTH_TYPE
# <b>content_length</b>::    CONTENT_LENGTH
# <b>content_type</b>::      CONTENT_TYPE
# <b>from</b>::              HTTP_FROM    
# <b>gateway_interface</b>:: GATEWAY_INTERFACE
# <b>path_info</b>::         PATH_INFO
# <b>path_translated</b>::   PATH_TRANSLATED
# <b>query_string</b>::      QUERY_STRING
# <b>raw_cookie</b>::        HTTP_COOKIE
# <b>referer</b>::           HTTP_REFERER
# <b>remote_addr</b>::       REMOTE_ADDR
# <b>remote_host</b>::       HTTP_HOST
# <b>remote_ident</b>::      REMOTE_IDENT
# <b>remote_user</b>::       REMOTE_USER
# <b>request_method</b>::    REQUEST_METHOD
# <b>script_name</b>::       SCRIPT_NAME
# <b>server_name</b>::       SERVER_NAME
# <b>server_port</b>::       SERVER_PORT
# <b>server_protocol</b>::   SERVER_PROTOCOL
# <b>server_software</b>::   SERVER_SOFTWARE
# <b>uri</b>, <b>url</b>::   REQUEST_URI
# <b>user_agent</b>::        HTTP_USER_AGENT
class CKRequest < CKMessage
	attr_accessor :multiple_params

	def initialize( headers, params )
		super headers
		@multiple_params = params || Hash.new([])
	end

	def single_param( key )
		if @multiple_params.key? key
			@multiple_params[key].join( ',' )
		else
			nil
		end
	end

	def keys
		@multiple_params.keys
	end
	
	def has_key? (key)
		@multiple_params.has_key? key
	end

	def cookie( key )
		@cookies.each { | cookie |
			if cookie.name == key
				return cookie
			end
		}
		nil
	end

	def cookie_value( key )
		@cookies.each { | cookie |
			if cookie.name == key
				return cookie.value
			end
		}
		nil
	end

	def cookie_values( key = nil )
		if key
			_cookie_values_for_key( @cookies, key )
		else
			_cookie_values @cookies
		end
	end

	alias [] single_param

	def accept;            self.headers['HTTP_ACCEPT']          end
	def accept_charset;    self.headers['HTTP_ACCEPT_CHARSET']  end
	def accept_language;   self.headers['HTTP_ACCEPT_LANGUAGE'] end
	def auth_type;         self.headers['AUTH_TYPE']            end
	def content_length;    self.headers['CONTENT_LENGTH']       end
	def content_type;      self.headers['CONTENT_TYPE']         end
	def from;              self.headers['HTTP_FROM']            end
	def gateway_interface; self.headers['GATEWAY_INTERFACE']    end
	def path_info;         self.headers['PATH_INFO']            end
	def path_translated;   self.headers['PATH_TRANSLATED']      end
	def query_string;      self.headers['QUERY_STRING']         end
	def raw_cookie;        self.headers['HTTP_COOKIE']          end
	def referer;           self.headers['HTTP_REFERER']         end
	def remote_addr;       self.headers['REMOTE_ADDR']          end
	def remote_host;       self.headers['HTTP_HOST']            end
	def remote_ident;      self.headers['REMOTE_IDENT']         end
	def remote_user;       self.headers['REMOTE_USER']          end
	def request_method;    self.headers['REQUEST_METHOD']       end
	def script_name;       self.headers['SCRIPT_NAME']          end
	def server_name;       self.headers['SERVER_NAME']          end
	def server_port;       self.headers['SERVER_PORT']          end
	def server_protocol;   self.headers['SERVER_PROTOCOL']      end
	def server_software;   self.headers['SERVER_SOFTWARE']      end
	def uri;               self.headers['REQUEST_URI']          end
	def user_agent;        self.headers['HTTP_USER_AGENT']      end

	alias url uri

	private

	def _cookie_values( cookies )
		values = {}
		cookies.each { | cookie |
			values[ cookie.name ] = cookie.value
		}
		values
	end

	def _cookie_values_for_key( cookies, key )
		values = []
		cookies.each { | cookie |
			if cookie.name == key
				values << cookie.value
			end
		}
		values
	end
end


# A response object that is sent to a browser by a CKAdapter object.
#
# == Getting a response object
# You can get a response object by CKApplication#response or CKComponent#response.
#
# == Setting headers for a response object
# To send HTTP response headers, append a pair of key and value to headers.
# For example,
#
#  application.response.headers['Content-Type'] = 'text/html'
#
# == Attributes
# <b>status</b>:: Status code in HTTP. Default status is 200 ( OK ).
class CKResponse < CKMessage
	attr_accessor :status

	STATUS = { 
		'100' => 'Continue',
		'101' => 'Switching Protocols',
		'200' => 'OK',
		'201' => 'Created',
		'202' => 'Accepted',
		'203' => 'Non-Authoritative Information',
		'204' => 'No Content',
		'205' => 'Reset Content',
		'206' => 'Partial Content',
		'300' => 'Multiple Choices',
		'301' => 'Moved Permanently',
		'302' => 'Found',
		'303' => 'See Other',
		'304' => 'Not Modified',
		'305' => 'Use Proxy',
		'307' => 'Temporary Redirect',
		'400' => 'Bad Request',
		'401' => 'Unauthorized',
		'402' => 'Payment Required',
		'403' => 'Forbidden',
		'404' => 'Not Found',
		'405' => 'Method Not Allowed',
		'406' => 'Not Acceptable',
		'407' => 'Proxy Authentication Required',
		'408' => 'Request Timeout',
		'409' => 'Conflict',
		'410' => 'Gone',
		'411' => 'Length Required',
		'412' => 'Precondition Failed',
		'413' => 'Request Entity Too Large',
		'414' => 'Request-URI Too Long',
		'415' => 'Unsupported Media Type',
		'416' => 'Requested Range Not Satisfiable',
		'417' => 'Expectation Failed',
		'500' => 'Internal Server Error',
		'501' => 'Not Implemented',
		'502' => 'Bad Gateway',
		'503' => 'Service Unavailable',
		'504' => 'Gateway Timeout',
		'505' => 'HTTP Version Not Supported'
	}

	def initialize( headers )
		super
		self.headers['Content-Type'] = 'text/html'
	end

	def to_s
		response = ''
		response << ( _status || '' )
		response << ( _general_header || '' )
		response << ( _response_header || '' )
		response << _entity_header
		response << EOL
		response << ( self.content || '' )
	end

	private

	def _status
		if @status then
			header =  "HTTP/#{self.http_version} #{self.status} "
			header << "#{STATUS[self.status]}"
			header << EOL
		end
	end

	def _general_header
		header = ''
		header << ( __header('Cache-Control') || '' )
		header << ( __header('Connection') || '' )
		header << ( __header('Date') || '' )
		header << ( __header('Pragma') || '' )
		header << ( __header('Trailer') || '' )
		header << ( __header('Transfet-Encoding') || '' )
		header << ( __header('Upgrade') || '' )
		header << ( __header('Via') || '' )
		header << ( __header('Warning') || '' )
	end

	def _response_header
		header = ''
		header << ( __header('Accept-Ranges') || '' )
		header << ( __header('Age') || '' )
		header << ( __header('ETag') || '' )
		header << ( __header('Location') || '' )
		header << ( __header('Proxy-Authenticate') || '' )
		header << ( __header('Retry-After') || '' )
		header << ( __header('Server') || '' )
		header << ( __header('Vary') || '' )
		header << ( __header('WWW-Authenticate') || '' )
	end

	def _entity_header
		header = ''
		header << ( __header('Allow') || '' )
		header << ( __header('Content-Encoding') || '' )
		header << ( __header('Content-Language') || '' )
		header << ( __header('Content-Length') || '' )
		header << ( __header('Content-Location') || '' )
		header << ( __header('Content-MD5') || '' )
		header << ( __header('Content-Range') || '' )
		header << __content_type
		header << ( __header('Expires') || '' )
		header << ( __header('Last-Modified') || '' )
		header << ( __cookie || '' )
	end

	def __header( header )
		"#{header}: #{self.headers[header]}#{EOL}" if self.headers[header]
	end

 	def __content_type
		header = "Content-Type: #{self.headers['Content-Type']}"
		header << "; charset=#{self.encoding}" if self.encoding
		header << EOL
	end

	def __cookie
		return if @cookies.empty?

		header = ''
		@cookies.each { | cookie |
			header << "Set-Cookie: #{cookie.to_s}"
			header << EOL
		}
		header
	end
end


# CKCookie is a class for cookie.
# To send cookies to a browser needs to create cookie objects 
# and set them to a response object. Instead of creating cookie objects,
# you can also get cookie objects from a request object.
#
# CKCookie objects have a pair of a cookie name and value.
# If you make the objects have multiple values for one name, 
# you must write code by yourself. 
# 
# == Controlling cookie objects
#
# === Creating cookies
# Give arguments of initialize() a name or a pair of name/value.
# The value of cookie is omittable.
#
#  cookie = CKCookie.new( name, value )
#
# === Getting cookies from a request object
# CKRequest has some methods for getting cookies.
# The methods are cookie(key), cookies, cookie_value(key), cookie_values(key).
# See also CKRequest.
#
# === Setting cookies to a response object
# CKResponse has methods for setting cookies. These methods are 
# defined in CKMessage, the superclass of CKResponse.
# Use add_cookie(cookie) and remove_cookie(cookie).
#
# == Attributes
# <b>name</b>::    Name of the cookie.
# <b>value</b>::   Value of the cookie.
# <b>path</b>::    Restricts the cookie in the site.
# <b>domain</b>::  Domain that can receive the cookie.
# <b>expires</b>:: Expiry date. You set Time object to the cookie object.
#                  The value is formatted when the cookie is returned.
# <b>secure</b>::  Decides whether the cookie is encrypted or not.
class CKCookie
	attr_accessor :name, :value, :path, :domain, :expires, :secure

    def initialize( name, value = nil )
		@name  = name
		@value = value
	end

	def value=( value )
		@value = value.to_s
	end

	def to_s
		buf = "#@name="
		if @value          then buf << CKUtilities.escape_url(@value) end
		if @domain         then buf << "; domain=#@domain" end
		if @path	       then buf << "; path=#@path" end
		if @expires        then
			buf << "; expires=#{CKUtilities.date(@expires)}"
		end
		if @secure == true then buf << '; secure' end
		buf
	end
end


# CKDefinition parses .ckd ( CGIKit Definition ) file.
# CKComponent stores the result of parse
# and uses it for binding.
#
# The format of CKD file is very different from Ruby or HTML
# Here, the format of CKD file is described as pseudo-BNF like Ruby's document.
#
#   STMTS           : (STMS)*
#   STMT            : ELEMENT_NAME `:' ELEMENT_TYPE `{' ATTRS `}'
#   ATTRS           : (ATTRIBUTE `=' VALUE)*
#   ATTRIBUTE       : ruby_method_name
#   VALUE           : METHODS `;'
#                   | LITERAL `;'
#   METHODS         : ruby_method_name(`.'ruby_method_name)*
#   LITERAL         : `true' 
#                   | `false' 
#                   | DIGIT_SEQUENCE 
#                   | STRING 
#   DIGIT_SEQUENCE  : DIGIT+
#   DIGIT           : `0' | `1' | `2' | `3' | `4' | `5' | `6' | `7' | `8' | `9'
#   STRING          : `'' any_character* `''
#                   | `"' any_character* `"'
#   ELEMENT_NAME    : (LETTER | DIGIT | `_')+
#   ELEMENT_TYPE    : ruby_class_name
#   LETTER          : `A' | `B' | `C' | `D' | `E' | `F' | `G' | `H' | `I' | `J' | `K' | `L' | `M' | `N' | `O' | `P' | `Q' | `R' | `S' | `T' | `U' | `V' | `W' | `X' | `Y' | `Z' | `a' | `b' | `c' | `d' | `e' | `f' | `g' | `h' | `i' | `j' | `k' | `l' | `m' | `n' | `o' | `p' | `q' | `r' | `s' | `t' | `u' | `v' | `w' | `x' | `y' | `z'
#
# ruby_method_name means Ruby's Method Name: For example, `foo', `to_s' and `is_a?'.
# Also, ruby_class means Ruby's Class Name: For instance, `CKString' and `CKComponent'.
#
# CKDefinition regards a part which begins from `#' to the end of the line as comment.
# In the next case, `# foo bar comment' is not interpreted by CKDefinition.
#
#   foo : CKString {
#     # foo bar comment
#     value = "foo";
#   }
#
class CKDefinition
	class << self
		def parse_ckd_file( file )
			parse_ckd_string ( _read_ckd_file file )
		end

		def parse_ckd_string( string )
			string      = _delete_comment string
			definitions = {}
			definition  = string.split '}'
			line        = ''

			definition.each { | line |
				myclass, mystruct = line.split '{'

				myclass = myclass.delete " \t\r\n"
				myname, myelement = myclass.split ':'
				definitions[ myname ] = {}
				definitions[ myname ][ 'oid' ]     = myname
				definitions[ myname ][ 'element' ] = myelement

				if mystruct == nil then next end
				mysyntax = mystruct.split ';'

				line2 = ''
				mysyntax.each { | line2 |
					next if line2 =~ /^\#/

					key, value = line2.split '='
					if not ( key and value ) then next end

					key = key.delete " \t\r\n"
					if value =~ /"(.*)"/
						value = "\"#$1\""
					elsif value =~ /'(.*)'"/
						value = "\"#$1\""
					else
						value = value.delete " \t\n"
					end
					definitions[ myname ][ key ] = value
				}
			}

			definitions
		end

		private

		def _read_ckd_file( file )
			f = File.new file
			string = f.readlines.join
			f.close
			string
		end

		def _delete_comment( ckd_string )
			string = ''
			ckd_string.each_line { | line |
				string << line.gsub( /\#(.*)/, '' )
			}
			string
		end
	end
end


# CKHTMLParser parses a template of a component.
# The template is the same as HTML, except one fact that it has
# "<cgikit>" tag. You can use two types of formats as CGIKit tag,
# "<cgikit>...</cgikit>" and "<cgikit />". The difference of the
# formats is whether the tag's body is empty or not. So, if you write
# "<cgikit></cgikit>", it has the same meaning as "<cgikit />".
#
# The CGIKit tag has only one attribute, "name". The "name" attribute
# is a name of CGIKit element. The "name" attribute is composed of "[a-zA-Z0-9_]".
# You can enclose the value of the "name" attribute by double quotation mark.
# And, The value is case-sensitive. For instance, these have the same meaning.
#
#   <cgikit name=foo></cgikit>
#   <cgikit name=foo />
#   <cgikit name="foo"></cgikit>
#   <cgikit name="foo" />
#
# On the other hand, these have different meanings.
#   <cgikit name=Foo />
#   <cgikit name="foo" />
#   <cgikit name=FOO></cgikit>
#
# Comment of HTML is interpreted by CKHTMLParser because it is
# necessary to include something like CSS and Javascript.
class CKHTMLParser
	attr_accessor :application, :html_string, :parent

	def initialize( parent, string )
		@parent      = parent
		@application = @parent.application
		@html_string = string
	end

	def parse
		@_tmp        = @html_string
		@_array      = []
		@_index      = 0
		string = body = tag = ""
		count  = 0
		flag   = false

		_parse_char
		while tag = _next
			if ( _cgikit_tag?( tag ) == true ) or ( flag == true ) \
				and ( _comment_tag?( tag ) == false ) then
				if flag == false then
					body = ''
					name = _name_cgikit_tag( tag )
				end
				flag = true
				body << tag

				if _cgikit_tag?( tag ) then
					if _concluded_tag? tag
						# do nothing
					elsif _open_tag? tag
						count += 1
					else
						count -= 1
					end
				end

				if count == 0 then
					flag    = false
					element = _create_element( name, _body_cgikit_tag( body ) )
					if parsed = element.to_s then string << parsed end
					body = ""
				end
			else
				string << tag
			end
		end
		string
	end
	
	private
	
	def _parse_char
		string = char = ""
		is_tag = false
		is_comment = false

		@_tmp.scan( /./m ) { | char |
			if ( char == "<" ) and ( is_comment == false )
				is_tag = true
				@_array.push( string )
				string = char
			elsif ( char == ">" ) and ( is_tag == true )
				string << ">"
				is_comment = false if string =~ /-->$/
				if is_comment == false
					@_array.push( string )
					is_tag = false
					string = ""
				end
			else
				string << char
				if ( string =~ /^<!--/ ) and ( is_comment == false )
					is_comment = true
				end
			end
			
			prechar = char
		}
		@_array.push( string )
	end
	
	def _next
		@_index += 1
		return @_array[ @_index - 1 ]
	end
	
	def _cgikit_tag?( tag )
		atts = tag.split( " " )
		if atts[0] == nil then return false end
		name = atts[0]
		if ( ( name =~ /<cgikit/i ) or ( name =~ /<\/cgikit/i ) ) then
			return true
		else
			return false
		end
	end
	
	def _concluded_tag?( tag )
		if ( tag =~ /<cgikit/i ) and ( tag =~ /\/>$/ )
			true
		else
			false
		end
	end

	def _open_tag?( tag )
		if tag =~ /<\/cgikit/i
			false
		else
			true
		end
	end
	
	def _name_cgikit_tag( tag )
		atts = tag.split( " ", 2 )
		if atts[0] == nil then return false end
		
		( key, value ) = atts[1].split( "=" )
		value.chop!
		value.tr!("\"\'/\s",'')
		value
	end
	
	def _body_cgikit_tag( body )
		body =~ /<cgikit name=(.*?)>(.*)<\/cgikit>/im
		tagbody = $2
	end

	def _comment_tag?( tag )
		if tag =~ /^<!--/
			true
		else
			false
		end
	end
	
	def _create_element( name, body )
		if parent.definitions[ name ] == nil then return nil end

		element = parent.definitions[ name ][ 'element' ]
		object = CKElement.instance( element, @application, @parent, name, body )
		object.run
		object
	end
end

# CKUtilitis is a class wihch collects utility methods based on cgi.rb.
class CKUtilities
	CR            = "\015"
	LF            = "\012"
	EOL           = CR + LF
	RFC822_DAYS   = %w[ Sun Mon Tue Wed Thu Fri Sat ]
	RFC822_MONTHS = %w[ Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ]

	class << self
		# Returns an encoded string for URL.
		def escape_url( string )
			string.to_s.gsub( /([^ a-zA-Z0-9_.-]+)/n ) do
				'%' + $1.unpack( 'H2' * $1.size ).join( '%' ).upcase
			end.tr( ' ', '+' )
		end

		# Returns a string decoded from URL.
		def unescape_url( string )
			string.to_s.tr( '+', ' ' ).gsub( /((?:%[0-9a-fA-F]{2})+)/n ) do
				[ $1.delete( '%' ) ].pack( 'H*' )
			end
		end

		# Escapes HTML control characters.
		def escape_html( string )
			string.to_s.gsub(/&/n, '&amp;').gsub(/\"/n, '&quot;').
				gsub(/>/n, '&gt;').gsub(/</n, '&lt;')
		end

		# Unescapes HTML control characters.
		def unescape_html( string )
    		string.to_s.gsub(/&(.*?);/n) do
		    	match = $1.dup
		    	case match
	    		when /\Aamp\z/ni           then '&'
				when /\Aquot\z/ni          then '"'
				when /\Agt\z/ni            then '>'
				when /\Alt\z/ni            then '<'
				when /\A#0*(\d+)\z/n       then
					if Integer($1) < 256
						Integer($1).chr
					else
						if Integer($1) < 65536 and \
							($KCODE[0] == ?u or $KCODE[0] == ?U)
							[Integer($1)].pack("U")
						else
							"&##{$1};"
						end
					end
				when /\A#x([0-9a-f]+)\z/ni then
					if $1.hex < 256
						$1.hex.chr
					else
						if $1.hex < 65536 and \
							($KCODE[0] == ?u or $KCODE[0] == ?U)
							[$1.hex].pack("U")
						else
							"&#x#{$1};"
						end
					end
				else
					"&#{match};"
				end
			end
		end

		# Formats Time object in RFC1123.
		# For example, "Sat, 1 Jan 2000 00:00:00 GMT".
		def date( time )
			t = time.clone.gmtime
			return format("%s, %.2d %s %.4d %.2d:%.2d:%.2d GMT",
				RFC822_DAYS[t.wday], t.day, RFC822_MONTHS[t.month-1], t.year,
				t.hour, t.min, t.sec)
		end

		def parse_query_string( query )
			params = Hash.new([])
			query.split(/[&;]/n).each do |pairs|
				key, value = pairs.split('=',2).collect{|v| unescape_url(v) }
				if params.has_key?(key)
					params[key].push(value)
				else
					params[key] = [value]
				end
			end
			params
		end

 		def parse_cookie( raw_cookie )
			cookies = []
			return cookies unless raw_cookie

			raw_cookie.split('; ').each do |pairs|
				name, value = pairs.split('=',2)
				name  = unescape_url name
				value = unescape_url value

				cookies << CKCookie.new( name, value )
			end

			cookies
		end
	end
end


# CKLog is a simple logging class.
class CKLog
	class << self
		# Outputs a debug string to STDERR.
		# A backtrace is sent if "exception" is given an error object.
		def debug( string, exception = nil )
			$stderr.write _debug_string(string,exception)
		end

		# Outputs a debug string to a file.
		def debug_to_file( file, string, exception = nil )
			CKFileLock.exclusive_lock( file ) { | out |
				out.write _debug_string(string,exception)
			}
		end

		# Outputs a debug string to "out" needs to be an IO object.
		def debug_to_out( out, string, exception = nil )
			out.write _debug_string(string,exception)
		end

		private

		def _debug_string( string, exception )
			str = "[#{Time.new.localtime}] #{CKApplication.version}: #{string}\n"
			if exception
				str << exception.backtrace.join << "\n"
			end
			str
		end
	end
end


# The super class for locking.
class CKLock
end

# CKFileLock is for locking files.
class CKFileLock < CKLock
	# Creates a shared file lock on a file.
	def self.shared_lock(filename)
		File.open(filename, 'rb') do |io|
			io.flock(File::LOCK_SH)
			yield io
			io.flock(File::LOCK_UN)
		end
	end

	# Creates a exclusive file lock on a file.
	def self.exclusive_lock(filename)
		File.open(filename, 'wb+') do |io|
			io.flock(File::LOCK_EX)
			yield io
			io.flock(File::LOCK_UN)
		end
	end
end


module CKAdaptee # :nodoc:
class CGI
	attr_accessor :headers, :params
	attr_reader :input, :output, :error

	def initialize
		@headers = ENV
		@input   = $stdin
		@output  = $stdout
		@error   = $stderr
		_initialize_query
	end

	def request
		CKRequest.new( headers, params )
	end

	def response
		CKResponse.new headers
	end

	def out( response )
		puts response
	end

	protected

	# Copied from cgi.rb.
	# Other interfaces should override _parse* methods.
    def _initialize_query
		if ("POST" == @headers['REQUEST_METHOD']) and
			 %r|\Amultipart/form-data.*boundary=\"?([^\";,]+)\"?|n.match(@headers['CONTENT_TYPE'])
			_parse_multipart
		else
			case @headers['REQUEST_METHOD']
			when "HEAD" then query = _parse_head
			when "GET"  then query = _parse_get
			when "POST" then query = _parse_post
			else             query = _parse_else end
			@params = CKUtilities.parse_query_string query
		end
	end

	def _parse_multi_part
		boundary = $1.dup
		@params = read_multipart(boundary, Integer(@headers['CONTENT_LENGTH']))
	end

	def _parse_get
		@headers['QUERY_STRING'] or ""
	end

	alias :_parse_head :_parse_get

	def _parse_post
		@input.binmode
		@input.read(Integer(@headers['CONTENT_LENGTH'])) or ''
	end

	def _parse_else
		require "shellwords"

		string = unless ARGV.empty?
			ARGV.join(' ')
		else
			if STDIN.tty?
				STDERR.print(
					%|(offline mode: enter name=value pairs on standard input)\n|
				)
			end
			readlines.join(' ').gsub(/\n/n, '')
		end.gsub(/\\=/n, '%3D').gsub(/\\&/n, '%26')

		words = Shellwords.shellwords(string)

		if words.find{|x| /=/n.match(x) }
			words.join('&')
		else
			words.join('+')
		end
	end
end
end

