# The super class of CGIKit error classes.
class CKError < StandardError; end


# This class is one of the most important classes in CGIKit.
# CKApplication has two major roles. 
# 
# One role is that CKAplication provides parameters which 
# effect the whole behavior of CGIKit application.
# The other role is that CKApplication communicates with CKAdapter, 
# an interface between a web server and CGIKit.
# 
# When CGIKit receives a request from a client,
# CKApplication arranges its parameters and creates a CKAdapter object.
# Next, CKApplication gets a request object from the CKAdapter object. 
# In reponse to the request object, CKApplication creates 
# a response object and sends it to the CKAdapter object.  
#
# == Application Process
#
# === 1. Ready a request and response object
# First, CKApplication creates a CKAdapter object and gets a request ( CKRequest ) 
# object from it. Then, CKApplication creates a response ( CKResponse ) 
# object. 
#
# === 2. Event loop
# Next, CKApplication goes into event loop. In this loop, CKApplication creates 
# components depending on the request and invokes the specified method of the components
# until it returns an object whose class is not CKComponent.
#
# === 3. Return the response object to the CKAdapter object
# Finally, CKApplication converts the components created in event loop to HTML
# and adds the HTML to the CKResponse object created in 1.
# CKApplication sends the response object to the CKAdapter object.
# Then, the CKAdapter object displays the response object to a client.
# 
# == CKApplication Attributes
# 
# === Controlling application
# <b>target</b>::         Component to be displayed.
# <b>action</b>::         Invoked method name of the target component.
# <b>main</b>::           Main component. If target isn't specified,
#                         this component is shown.
#                         The default value is 'MainPage'.
# <b>path</b>::           Path of an application.
# <b>component_path</b>:: Path for components.
#                         Components are searched under this path.
# <b>document_root</b>::  Document root directory.
# <b>locale</b>::         Locale of an application in a transaction.
#                         If the locale is specified, CGIKit change
#                         a template for a component. The name of
#                         the template includes the component name and
#                         the locale name. Also, the template name is 
#                         seprated by underscore("_"). For example,
#                         if the locale is "ja" and
#                         the component name is "MainPage",
#                         the template name is "MainPage_ja.html".
# <b>master_locale</b>::  Main locale of an application in a transaction.
#                         If the master_locale is equal to the specified
#                         locale, the components use templates whose name
#                         doesn't include the locale name. For instance,
#                         if the master_locale is "ja", the locale is "ja"
#                         and the component name is "MainPage",
#                         the template name is "MainPage.html".
# <b>resource</b>::       Resource directory. If "file" attribute of
#                         CKImage is set, the CKImage object searches
#                         an image file from resource directory. ( classes
#                         which manages other resources will be developed. )
# <b>interface</b>::      Interface of adapter.
#                         The default value is an interface for CGI.
#
# === Debugging
# <b>debug</b>::            Shows the backtrace in an error page
#                           when true is set.
# <b>attributes_check</b>:: Checks required attributes of elements
#                           if true is set.
#                           If undefined attributes are used or
#                           required attributes don't exist,
#                           an error page is displayed when the value is true.
class CKApplication
	attr_accessor :target, :action, :main, :locale, :master_locale, \
		:baseurl, :document_root, :path, :component_path, :resource, \
		:adapter, :interface, :request, :response, :debug, :attributes_check

	VERSION          = 'CGIKit 1.0b5'
	PATH             = Dir.pwd
	COMPONENT_PATH   = Dir.pwd
	MAIN_PAGE        = 'MainPage'
	ATTRIBUTES_CHECK = true
	ADAPTER          = 'cgi'
	ERROR_PAGE       = 'CKErrorPage'

	class << self
		# Returns version of CGIKit.
		def version
			VERSION
		end
	end

	def initialize
		@path             = PATH
		@component_path   = COMPONENT_PATH
		@main             = MAIN_PAGE
		@attributes_check = ATTRIBUTES_CHECK
	end

	# Returns the name of the application without file extension.
	def name
		basename = File.basename( @path, '.*' )
	end

	def master_locale?
		if @locale == @master_locale then
			true
		elsif not @locale then
			true
		elsif ( ( not @master_locale ) and ( not @locale ) ) then
			true
		else
			false
		end
	end

	# This method runs the CGIKit application.
	#
	#  require 'cgikit'
	#
	#  app = CKApplication.new
	#  app.run
	def run
		begin
			_ready_request_response
			component = _event_loop
			@response.content = component.to_s
		rescue Exception => e
			@response = handle_error e
		end

		@adapter.run @response
	end

	def create_adapter
		unless self.interface then
			CKAdapter.new ADAPTER
		else
			CKAdapter.new self.interface
		end
	end

	def handle_error( error )
		error_page       = page ERROR_PAGE
		error_page.error = error
		error_page.debug = @debug
		error_page.create_response		
	end

	def page( name )
		CKElement.instance( name, self, nil, nil, nil )
	end

	private

	def _ready_request_response
	 	@adapter = create_adapter unless @adapter

		# Gets request and creates response.
		@request        = @adapter.create_request  unless @request
		@response       = @adapter.create_response unless @response
		@target         = @request['target']
		@action         = @request['action']
		@locale         = @request['locale']
		@master_locale  = @request['master_locale']
		@baseurl        = @request.script_name unless @baseurl
		@document_root  = @request.headers['DOCUMENT_ROOT']  \
			unless @document_root
		@target = @main.dup unless @target
	end

	def _event_loop
		result = page @target

		# Recreates a main component if the component includes CKPartsMaker.
		if ( result.is_a? CKPartsMaker ) and ( @action == nil )
			@target = result.substitute_page || @main
			result  = page @target
		end

		begin
			component = result
			result    = component.run
		end while result.is_a? CKComponent

		component
	end
end


# CKBinding.bind invokes a method ( action ) of a target ( object ).
# You can get the result of the method by CKBinding.bind like Kernel#send.
# The method call can accompany arguments.
#
# == Order of method search  is as follows
#
# === Getting methods
# 1. object.field
# 2. object.get_field
# 3. object.getField
# 4. object._field
# 5. object._get_field
# 6. object._getField
#
# === Setting methods
# 1. object.field=
# 2. object.set_field
# 3. object.setField
# 4. object._field=
# 5. object._set_field
# 6. object._setField
class CKBinding
	class << self
		def bind( target, action, *args )
			if not ( target and action ) then return nil end
			_object, _getOrder, _setOrder = _orders( target, action )

			# make decidion of accessors
			result = nil
			if   args.empty? then order = _getOrder
			else                  order = _setOrder end

			is_respond = false
			order.each { | method |
				if method =~ /\"(.*)\"/
					result = $1
				elsif _object.respond_to?( method ) == false
					next
				end

				if result then
					is_respond = true
					break
				elsif args.empty? then
					result     = _object.send method
					is_respond = true
					break
				else
					result     = _object.send( method, *args )
					is_respond = true
					break
				end
			}

			if is_respond == false
				raise NameError, \
					"No such variable or method - #{target.class}\##{action}"
			end
			result
		end

		private

		def _orders( target, action )
			_object = target

			if action.index( '.' ) then
				methods = action.split( '.' )
				message = methods.pop
				methods.each { | method |
					_object = _object.__send__( method )
				}
			else
				message = action
			end
		
			accessor = message[ 0, 1 ].upcase + message.slice( 1, action.size )
			_getOrder = [ message, "get_#{message}", "get#{accessor}", \
						  "_#{message}", "_get_#{message}", "_get#{accessor}" ]
			_setOrder = [ "#{message}=", "set#{accessor}", "set#{accessor}", \
						  "_#{message}=", "_set_#{message}", "_set#{accessor}" ]
			[ _object, _getOrder, _setOrder ]
		end
	end
end


# Creates URL to access components and methods.
class CKURL
	attr_accessor :secure, :target, :action, :oid, :query, :parent

	def initialize( application )
		@application = application
	end

	def url
		baseurl     = @application.baseurl
		locale      = @application.locale
		master      = @application.master_locale
		if target == nil then
			@target = @application.target
		end

		if ( secure and ( not baseurl =~ /^https:\/\//i ) ) then
			baseurl =~ /^http:\/\/(.*)/i
			if $1 then string = "https://#{$1}?"
			else       string = "https://#{baseurl}?" end
		else
			string = "#{baseurl}?"
		end

		if target       then string << "target=#{target}&" end
		if action       then string << "action=#{action}&" end
		if master       then string << "master_locale=#{master}&" end
		if locale       then string << "locale=#{locale}&" end
		if oid          then string << "oid=#{oid}&"       end
		if query        then string << query_string( query ) end
		string
	end

	def query_string( hash )
		str = ''
		hash.each { | key, value |
			str << "#{key}=#{CKUtilities.escape_url( value )}&"
		}
		str
	end
end

