# The super class of dynamic element classes.
#
# Dynamic elements convert themselves to HTML.
# These are very important in development with CGIKit.
#
# CKElement returns HTML when to_s() is called.
# In components development, you bind elements to component's methods. 
# At runtime, the binding is executed by run(). 
# These two methods are expected to be overrided in the subclasses of CKElement.
#
# == About CKElement#init_element
# In initialize(), CKElement calls init_element().
# You can hook the inialization of elements by this method.
# The reason why element classes have init_element() is for convenience.
# It may be better to override init_element() than initialize().
#
# * CKElement#initialize
#  class MainPage < CKComponent
#    def initialize( app, parent, name, body )
#      super( app, parent, name, body )
#
#      # initializing code ...
#
#    end
#  end
#
# * CKElement#init_element
#
#  class MainPage < CKComponent
#    def init_element
#
#      # initializing code ...
#
#    end
#  end
#
# == Paths for searching elements and components
# CKElement objects, including components, are usually instantiated
# by CKElement.instance. This method loads elements/components and creates
# element objects. In this method, some paths are searched to require files.
#
# The searched paths:
# 1. CKApplicaton#component_path
# 2. ($LOAD_PATH)/cgikit/elements
# 3. ($LOAD_PATH)/cgikit/components
#
# The latter two paths are for extention elements or components.
# It is recommended that your own elements or components are installed in the 1st path.
#
# == Creation of custom elements
# If you create custom elements, the class of the custom elements must 
# inherit CKElement. Simple dynamic element can be created only 
# by overriding to_s(). to_s() is called when CGIKit
# converts the dynamic elements to HTML. If you add new attributes
# to your custom elements, run() must be overrided. 
class CKElement
	class UnknownElementError < CKError; end
	class AttributeError      < CKError; end

	attr_accessor :name, :body, :definition, :parent, :application

	class << self
		def instance( element, app, parent, name, body )
			body          = '' unless body
			component     = "#{app.component_path}/#{element}/#{element}"
			ext_element   = "cgikit/elements/#{element}"
			ext_component = "cgikit/components/#{element}/#{element}"
			paths         = [ ext_element, ext_component ]

			# starndard elements
			if Object.const_defined? element
				return Object.const_get(element).new(app,parent,name,body)
			# components
			elsif FileTest.exist? "#{component}.rb"
				require component
				obj = Object.const_get(element).new(app,parent,name,body)
				return obj
			end

			# check paths for extension components
			$LOAD_PATH.each { | load_path |
				paths.each { | path |
					ext_path = "#{load_path}/#{path}"
					if FileTest.exist?( "#{ext_path}.rb" )
						require path
						obj = Object.const_get(element).new(app,parent,name,body)
						obj.path = ext_path
						return obj
					else
						next
					end
				}
			}

			raise UnknownElementError, "No such '#{element}' element or component."
		end
	end

	def initialize( app, parent, name, body )
		self.application = app
		self.parent      = parent
		self.name        = name
		self.body        = body
		init_element
	end

	def init_element; end
	
	def definition
		parent.definitions[name]
	end

	def name_with_class
		"\"#{name}\":#{self.class}"
	end

	# Returns value of "name" attribute for form.
	def oid
		if self.class.to_s == 'CKSubmitButton' then
			"#{parent.class}.oid.#{name}"
		else
			"#{parent.class}.#{name}"
		end
	end
	
	# Fetches the result of the parent component's methods.
	# When is_binding is true, fetch() returns String without binding from . 
	def fetch( key, is_binding = true )
		data = definition[ key ]
		if is_binding == false
			data
		else
			_parse_ckd_value( parent, data )
		end
	end
	
	# This method will be overrided in subclasses.
	def run; end

	# Raises exception if an element has undefined attributes.
	def check_undefined_attributes( *attrs )
		attrs << 'oid'     unless attrs.include? 'oid'
		attrs << 'element' unless attrs.include? 'element'

		definition.each_key { | attr |
			if ( attrs.include? attr ) == false
				raise AttributeError, \
					"Attribute '#{attr}' is undefined attribute - #{name_with_class}"
			end
		}
	end

	# Raises exception if an element doesn't have required attributes.
	def check_required_attributes( *patterns )
		requires = 0
		attrs    = []

		patterns.each { | pattern |
			pattern.each { | attr |
				requires += 1 if definition.include? attr
			}

			return true if requires == pattern.size
			requires = 0
			attrs << ( "'" + pattern.join(', ') + "'" )
		}
		message =  "#{self.class} requires #{attrs.join(' or ')} "
		message << "of attribute - #{name_with_class}"
		raise AttributeError, message
	end
	
	private
	
	def _parse_ckd_value( target, string )
		if    string == 'true'         then value = true
		elsif string == 'false'        then value = false
		elsif string =~ /"(.*)"/       then value = $1
		elsif string =~ /'(.*)'/       then value = $1
		elsif ( ( not string =~ /[\D]/ ) and ( string =~ /[\d]/ ) ) then
			value = string.to_i
		else
			value = CKBinding.bind( target, string )
		end
		value
	end
end


# When you define your component, its class must inherit CKComponent in code
# file. The component dynamically renders HTML with template and binding file. 
#
# == Construction of component
# A component has 3 files in a directory whose name is the same as the
# component's name. These three files composite Model-View-Controller
# architecture.
#
# Template(View)::      A HTML file includes CGIKit tags ( <cgikit> ). One
#                       CGIKit tag corresponds to one dynamic element.
#
# Binding(Controller):: A definition file for elements is used by a component.
#                       The suffix, ckd, is abbreviation of "CGIKit Definition".
#                       The definition file connects Template to Code.
#
# Code(Model)::         A Ruby source where a component's class is defined. 
#                       It must inherit CKComponent. The name of the component's
#                       class must be the same as the component name. 
#
# ex) An application with a MainPage component
#  /cgi-bin/cgikit-app
#    cgikit-app.cgi
#    /MainPage
#      MainPage.html
#      MainPage.ckd
#      MainPage.rb
#
# In this case, CGIKit application is cgikit-app.cgi and one component,
# "MainPage". 
#
# === Template(View)
# Template is an ordinary HTML file, except CGIKit's own tag. 
# CGIKit tag has only one attribute, "name". Each <cgikit> tag in template 
# corresponds to one dynamic element.
#
# (MainPage.html)
#  <html>
#  <body>
#  <cgikit name="HelloWorld"/>
#  </body>
#  </html>
#
# === Binding(Controller)
# In a binding file, a type and attributes of the elements in the template
# are defined. Note that the format differs form Ruby.
#
# (MainPage.ckd)
#  HelloWorld : CKString {
#    value = sayHello;
#  }
# 
# ==== 1. ElementName : ElementType
# A name and type of the element must be separated with colon.
# In the example above, the element name is HelloWorld and 
# the element type is CKString. Note that HelloWorld is
# used in the template as "name" attribute of the CGIKit tag.
#
# ==== 2. {
# The blanket, "{", means beginning of the definition of the element attributes.
#
# ==== 3. attribute = value;
# In attributes defintion, you write some pairs of an element's attribute
# and its value. The value is method name or literal. 
# At the end of the pair, you must locate a semicolon.
# If you write string literal, you must enclose the string 
# with single or double quotation marks.
#
# ex)
# method::               <tt>attribute = method_name;</tt>
# string::               <tt>attribute = "string";</tt> or <tt>'string'</tt>
# figure(integer only):: <tt>attribute = 12345;</tt>
# true/false::           <tt>attribute = true;</tt>
#
# ==== 4. }
# Close the definition with a close-blanket.
#
# === Code(Model)
# A class of a component must be named the same name 
# as the component name, and inherit CKComponent.
# The class has the methods specified in the ckd file.
#
# (MainPage.rb)
#  class MainPage < CKComponent
#    def sayHello
#      "Hello World!"
#    end
#  end
#
# === Result
#  Hello World!
#
# In the HelloWorld example, the component is parsed in this order.
#
# 1. (Template) <cgikit name="HelloWorld"/>
# 2. (Binding) HelloWorld : CKString
# 3. (Binding) value = sayHello
# 4. (Code) MainPage#sayHello
# 5. (Code) "Hello World!"
# 6. (Browser) Hello World!
#
# == Calling methods of a component
# Some elements, for example, CKForm, CKSubmit and CKHyperlink,
# can call a method of a component when they receive some events.
# You can specify the called method by their "action" attribute.
#
# For example, if you bind Echo#echo method to a CKHyperlink object and
# the link created by the CKHyperlink object is clicked, Echo#echo is invoked.
# Please try running examples in samples/Examples by yourself.
#
# === Linking to other components
# Now, you can hook some events. Then, what can we do in the hooked event?
# When a submit button or hyperlink is clicked, a method bound to the 
# "action" attribute is called. If the result of the method is 
# a CKComponent object, the returned object is displayed as web page.
# If the returned object is not a CKComponent object, 
# the current component is displayed.
#
# CGIKit provides the easy way to create component objects.
# A method that creates an instance of component is page().
# It returns a object of the component specified by the name.
# You can link to other components by using this method.
#
# See example:
#
#  def other_page
#    # Create an instance of a component
#    other_page = page( 'OtherComponentName' )
#
#    # Control the object if needed
#    other_page.foo = true
#
#    # Return the component object
#    other_page
#  end
#
# This example displays a component, "OhterComponentName", 
# when a user clicks the element to which the method, "other_page", is bound.
# In this example, a OtherComponentName's object is created by page().
# After that, true is assigned to the "foo" attribute of the created object.
#
# Web page, the component object, can be controlled before returning to clients
# in this procedure.
#
# == Processing form data
# Normally, CGI scripts process form data by themselves.
# But, CGIKit applications don't have to do it.
#
# For example, when a web page is created from a template
# which has CKText, the "name" attribute of the input tag ( textarea ) is
# determined by the CKText. The value of the "name" attribute
# includes the ID. If a user clicks the submit button, the data of
# the input tag is assinged to the component by calling the
# method which the ID specifies. The method which the ID specifies
# is bound the CKText element to.
#
# <b>Note</b>: IDs are not unique. Its are created by joining names of
# components and elements that defined in the components.
# Don't use a defined element over and over. 
#
# == Nesting components
# You can use one component in another component like elements. It means that  
# it is possible to nest components. 
#
# === Parts maker
# You can write a component which converts itself to a
# part of a web page. In some cases, a web page is composed of these components.
# These components are called parts component. Parts componente
# is recommended to include CKPartsMaker. A component which includes
# CKPartsMaker isn't displayed even if its name is set to
# the CKApplication#target. See also CKPartsMaker section.
#
# === Example
# This is an example.  
#
# * Template of a nested component
#  <cgikit name="String"/>
#
# * Binding of a nested compoent
#  String : CKString {
#    value = string;
#  }
#
# * Code of a nested component
#  class NestedComponent < CKComponent
#    include CKPartsMaker
#    attr_accessor :string
#  end
#
# * Template of parent's component
#  <html>
#  ...
#  <cgikit name=Location>
#  ...
#  </html>
#
# * Binding of parent's component
#  Location : NestedComponent {
#    string = "This is the nested component.";
#  }
#
# * Result
#  <html>
#  ...
#  This is the nested component.
#  ...
#  </html>
#  
# As you can see, the nested component is bound to the method
# of the parent component.
class CKComponent < CKElement
	class FileNotFoundError < CKError; end

	attr_accessor :action, :path

	def request
		application.request
	end

	def response
		application.response
	end

	def create_response
		response = CKResponse.new nil
		response.content = self.to_s
		response
	end

	def html_string
		return @html_string if @html_string

		begin
			f = File.new( _view )
			@html_string = f.readlines.join
			f.close
		rescue
			raise FileNotFoundError, "Can't read a template file - #{_view}"
		end

		@html_string
	end
	
	def definitions
		return @definitions if @definitions

		unless FileTest.exist? _controller
			msg = "Can't read a binding file - #{_controller} #{self.class}"
			raise FileNotFoundError, msg
		end

		@definitions = CKDefinition.parse_ckd_file _controller
	end

	def initialize( app, parent, name, body )
		@path = "#{app.component_path}/#{self.class}/#{self.class}"
		super( app, parent, name, body )
	end

	# Returns an instance of the new component.
	def page( page )
		CKElement.instance( page, application, parent, nil, nil )
	end

	# Invokes the action after binding component's variables to attributes
	# for definitions, values from request to component's variables.
	def run
		_bind_attribute
		_bind_request
		if @action == nil then return end

		if self.class.to_s == application.target
			CKBinding.bind( self, @action )
		end
	end

	def to_s
		parser = CKHTMLParser.new( self, html_string )
		parser.parse
	end
	
	private

	def _view
		if application.master_locale? then
			"#{path}.html"
		else
			"#{path}_#{application.locale}.html"
		end
	end
	
	def _controller
		"#{path}.ckd"
	end
	
	def _bind_attribute
		if parent == nil then return end
		return unless parent
		return unless definition
		defaults = { "name" => true, "oid" => true, "element" => true }

		definition.each { | key, value |
			if defaults[key] then next end
			CKBinding.bind( self, key, _parse_ckd_value(parent,value) )
		}
	end
	
	def _bind_request
		request = application.request

		_target = self
		request.multiple_params.each { | key, value |
			( object, variable, method ) = key.split( '.' )
			
			if key == 'action' then
				@action = value[0];
				next
			# for CKRadioButton
			elsif ( ( variable == nil ) and ( method == nil ) and \
				( definitions[ value[0] ] ) ) then

				# checked or selection
				if definitions[ value[0] ][ 'checked' ] then
					_action = definitions[ value[0] ][ 'checked' ]
					_args   = true
				elsif definitions[ value[0] ][ 'selection' ] then
					_action = definitions[ value[0] ][ 'selection' ]
					__action = definitions[ value[0] ][ 'value' ]
					_args   = CKBinding.bind( self, __action )
				end

				CKBinding.bind( _target, _action, _args )
				next
			elsif object != self.class.to_s then
				next
			end
			
			# for CKSubmitButton
			if method then
				@action = definitions[method]['action']
				next
			end
			
			# for CKCheckbox
			if definitions[variable]['checked'] then
				setVariable = definitions[variable]['checked']
				value = true
			elsif definitions[variable]['selection'] then
				setVariable = definitions[variable]['selection']
			# for CKPopUpButton, CKBrowser
			elsif definitions[variable]['selected'] then
				setVariable = definitions[variable]['selected']
			else
				setVariable = definitions[variable]['value']
			end
			_action = setVariable
			
			# binding data
			if value == true then
			elsif definitions[ variable ][ 'element' ] == 'CKBrowser' then
				value.delete ""
			elsif value.size < 2 then
				value = value[0]
			end

			CKBinding.bind( _target, _action, value )
		}
	end
end


# CKContent is used in nested components.
# This element tag in the template is replaced
# with a part of the template of its grandparent component.
#
# == Example
# * Template of a nested component ( This component is the parent of CKContent element.)
#  <b><cgikit name=Content /></b>
#
# * Binding of a nested component
#  Content : CKContent {}
#
# * Code of a nested component
#  class NestedComponent < CKComponent
#    include CKPartsMaker
#  end
#
# * Template of a grandparent component
#  foo
#  <cgikit name=Nested>This is the nested component.</cgikit>
#  bar
#
# * Binding of a grandparent component
#  Nested : NestedComponent {}
#
# * Result
#  foo
#  <b>This is the nested component.</b>
#  bar
#
# Another example is available in samples/Examples.
# In samples/Examples, CKContent element is used in HeaderFooter, which 
# defines the header and footer for all of the appication's pages.
#
# <b>Note</b>: You can locate only one CKContent element in the
# template of a given component.
#
# == Bindings
# CKContent doesn't have any attribute.
class CKContent < CKElement
	def run
		if application.attributes_check
			check_undefined_attributes nil
		end
	end

	def to_s
		unless parent.parent
			return
		end

		parser = CKHTMLParser.new( parent.parent, parent.body )
		parser.parse
	end
end


# CKFrame generates frame tag in HTML.
#
# == Bindings
# Required attributes: <b>page</b>, <b>src</b> or <b>value</b>.
#
# The priorities of "src", "page" and "value" is "src", 
# "page", "value". 
# 
# <b>name</b>::  "name" attribute of HTML's <frame> tag.
# <b>page</b>::  Name of component that supplies the content
#                for the frame.
# <b>src</b>::   You specify the URL or file for the frame. 
# <b>value</b>:: Method that supplies the content. The parent of this element
#                must have the specified method.
class CKFrame < CKElement
	def run
		@attr_name  = fetch( 'name' )  # String
		@attr_page  = fetch( 'page' )  # String
		@attr_src   = fetch( 'src' )   # String
		@attr_value = fetch( 'value' ) # CKComponent

		if application.attributes_check
			check_undefined_attributes( 'name', 'page', 'src', 'value' )
			check_required_attributes( ['page'], ['src'], ['value'] )
		end
	end

	def to_s
		if @attr_src then
		elsif @attr_page then
			url = CKURL.new application
			url.target = @attr_page
			@attr_src = url.url
		elsif @attr_value then
			url = CKURL.new application
			url.target = parent.class
			url.action = @attr_value
			@attr_src = url.url
		end
		
		"<frame name=\"#@attr_name\" src=\"#@attr_src\">"
	end
end

# CKString shows the result of binding as string.
#
# == Bindings
# Required attributes: <b>value</b>
#
# <b>value</b>::  Text to be displayed.   
# <b>escape</b>:: Escapes HTML control characters in "value" if the "escape"
#                 is true. The default value is true.
# <b>empty</b>::  Text that is substituted for "value"
#                 when the "value" is nil.
class CKString < CKElement
	def run	
		@attr_value  = fetch( 'value' )  # String
		@attr_escape = fetch( 'escape' ) # true or false
		@attr_empty  = fetch( 'empty' )  # String
		@attr_escape = true unless @attr_escape

		if application.attributes_check
			check_undefined_attributes( 'value', 'escape', 'empty' )
			check_required_attributes( ['value'], ['empty'] )
		end
	end

	def to_s
		if ( @attr_empty ) and ( @attr_value == nil ) then
			@attr_value = @attr_empty
		end
		if ( @attr_escape ) and ( @attr_value ) then
			@attr_value = CKUtilities.escape_html @attr_value
		end

		@attr_value.to_s
	end
end


# CKHyperlink generates a hypertext link.
#
# == Bindings
# Required attributes: <b>action</b>, <b>href</b> or <b>page</b>.
#
# <b>action</b>::   Method to be invoked when the link is clicked.
# <b>disabled</b>:: Generates a non-active link if the value is true.
# <b>href</b>::     You specify the URL to other web page directly.
#                   This attribute prevails over "action" or "page" attribute.
# <b>page</b>::     Name of component to display when the link is clicked.
# <b>string</b>::   Text of the link. If the body of CKHyperlink tag is
#                   not empty, the body is displayed. For example,
#                   if the template includes "<cgikit name=link>foo</cgikit>",
#                   this element shows "foo" as the link.
# <b>target</b>::   target attribute of HTML's <a> tag. 
# <b>secure</b>::   Appends "https://" to the URL if the value is true.
#                   The default value is false.
# <b>query</b>::    Hash as the query string. The value of "query" attribute is
#                   converted to string. Then, the string is added to append
#                   the URL.
class CKHyperlink < CKElement
	def run
		@attr_action   = fetch( 'action', false ) # CKComponent
		@attr_disabled = fetch( 'disabled' )      # true or false
		@attr_href     = fetch( 'href' )          # String
		@attr_page     = fetch( 'page' )          # String
		@attr_string   = fetch( 'string' )        # String
		@attr_target   = fetch( 'target' )        # String
		@attr_secure   = fetch( 'secure' )        # true or false
		@attr_query    = fetch( 'query' )         # Hash

		if application.attributes_check
			check_undefined_attributes( 'action', 'disabled', 'href', \
				'page', 'string', 'target', 'secure', 'query' )
			check_required_attributes( ['action'], ['href'], ['page'] )
		end
	end

	def to_s
		# action
		url = CKURL.new application
		if @attr_page then target = @attr_page
		else               target = parent.class.to_s end
		url.target  = target
		url.parent  = parent
		url.action  = @attr_action
		url.secure  = @attr_secure
		url.query   = @attr_query if @attr_query
		@attr_action = url.url

		# to_s
		to_s = '<a'
		if   @attr_href    then to_s << " href=\"#@attr_href\""
		else                    to_s << " href=\"#@attr_action\""   end
		if   @attr_target  then to_s << " target=\"#@attr_target\"" end
		to_s << '>'
		if    @attr_string then to_s << @attr_string
		elsif body         then to_s << body end
		to_s << '</a>'
		
		if @attr_disabled == true then to_s = body end
		to_s
	end
end


# Creates an image tag.
#
# == Bindings
# Required attributes: <b>file</b> or <b>src</b>.
#
# <b>alt</b>::    Alternative text to the picture.
# <b>border</b>:: Size of image border.
# <b>file</b>::   Name of an image file searched from 
#                 resource directory(CKApplication#resource).
# <b>src</b>::    You specify an image file directly. This value prevails 
#                 over "file" attribute.
class CKImage < CKElement
	def run	
		@attr_alt    = fetch( 'alt' )    # String
		@attr_border = fetch( 'border' ) # Integer
		@attr_file   = fetch( 'file' )   # String
		@attr_src    = fetch( 'src' )    # String

		if application.attributes_check
			check_undefined_attributes( 'alt', 'border', 'file', 'src' )
			check_required_attributes( ['file'], ['src'] )
		end
	end

	def to_s
		if @attr_file then
			resource = application.resource || application.baseurl
			@attr_src = File.join( resource, @attr_file )
		end

		to_s = '<img'
		if @attr_src    then to_s << " src=\"#@attr_src\""       end
		if @attr_alt    then to_s << " alt=\"#@attr_alt\""       end
		if @attr_border then to_s << " border=\"#@attr_border\"" end
		to_s << '>'
	end
end


# Controls generating HTML.
#
# == Bindings
# Required attribute: <b>condition</b>
#
# <b>condition</b>:: If the value is true and "negate" is false, the body of the 
#                    CKCoditional tag is displayed.
# <b>negate</b>::    Inverts the meaning of the "condition".
#
# == Control table
#   condition  negate  result
#   true       false   show
#   false      false   not show
#   true       true    not show
#   false      false   show
class CKConditional < CKElement
	def run
		@attr_condition = fetch( 'condition' )
		@attr_negate    = fetch( 'negate' )

		if application.attributes_check
			check_undefined_attributes( 'condition', 'negate' )
			check_required_attributes( ['condition'] )
		end
	end

	def to_s
		to_s = ''
		if ( ( ( not @attr_negate    ) and @attr_condition ) or \
		     ( ( not @attr_condition ) and @attr_negate    ) ) then
			parser = CKHTMLParser.new( parent, body )
			to_s = parser.parse
		end
		to_s
	end
end


# Creates a fill-in form.
# Dynamic Elements of form, for example, CKBrowser, CKCheckbox, CKRadioButton,
# CKPopUpButton, CKText, CKTextField, CKSubmitButton and CKResetButton,
# are used within CKForm or HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>action</b>::  Method to invoke when a submit button in the form is clicked.
# <b>enctype</b>:: Encode type to send form data.
#                  You can use "POST", "GET" or "HEAD" as the value.
# <b>href</b>::    URL to which the browser directs.
# <b>target</b>::  Frame in a frameset that receive the page.
class CKForm < CKElement
	def run
		@attr_action  = fetch( 'action' )  # String
		@attr_enctype = fetch( 'enctype' ) # String
		@attr_href    = fetch( 'href' )    # String
		@attr_target  = fetch( 'target' )  # String
		if ( application.baseurl == nil ) and ( @attr_href == nil )
			@attr_href = application.request.script_name
		end
		if @attr_enctype == nil then @attr_enctype = 'POST' end
		unless ( @attr_enctype.upcase == 'POST' ) or \
			( @attr_enctype.upcase == 'GET' ) or \
			( @attr_enctype.upcase == 'HEAD' )
			@attr_enctype = 'POST'
		end

		if application.attributes_check
			check_undefined_attributes( 'action', 'method', 'enctype', \
				'href', 'target' )
		end
	end

	def to_s
		parser = CKHTMLParser.new( parent, body )
		value = parser.parse
		
		to_s = "<form method=\"#@attr_enctype\""
		if @attr_href    then to_s << " action=\"#@attr_href\""
		else                  to_s << " action=\"#{application.baseurl}\"" end
		if @attr_target  then to_s << " target=\"#@attr_target\"" end
		if @attr_enctype then to_s << " enctype=\"#@attr_enctype\"" end
		to_s << ">\n"
		to_s << "<input type=\"hidden\" name=\"target\" value=\"#{parent.class}\">\n"
		if @attr_action then
			to_s << "<input type=\"hidden\" name=\"action\" value=\"#@attr_action\">"
		end
		to_s << "#{value}\n</form>\n"
	end
end


# Creates a list whose multiple items can be selected.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>list</b>
#
# <b>escape</b>::    Escapes HTML control characters in "value"
#                    if the "escape" is true. The default value is true.
# <b>list</b>::      Array which is iterated through. The first object
# <b>default</b>::   The first item if no item is selected. 
# <b>selected</b>::  Items which are chosen from the list.
# <b>multiple</b>::  Multiple items of the list can be selected 
#                    if the value is true.
# <b>size</b>::      Size of item in appearance.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKBrowser < CKElement
	def run
		@attr_name     = oid
		@attr_escape   = fetch( 'escape' )
		@attr_list     = fetch( 'list' )
		@attr_default  = fetch( 'default' )
		@attr_selected = fetch( 'selected' )
		@attr_multiple = fetch( 'multiple' )
		@attr_size     = fetch( 'size' )
		@attr_disabled = fetch( 'disabled' )
		if @attr_selected == nil then @attr_selected = [] end

		if application.attributes_check
			check_undefined_attributes( 'escape', 'list', \
				'default', 'selected', 'multiple', 'size', 'disabled' )
			check_required_attributes( ['list'] )
		end
	end

	def to_s
		to_s = '<select'
		if @attr_name     then to_s << " name=\"#@attr_name\"" end
		if @attr_size     then to_s << " size=\"#@attr_size\"" end
		if @attr_multiple then to_s << " multiple" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">\n"
		if @attr_default then
			to_s << "<option value=\"\""
			if @attr_selected == nil then
				to_s << " selected=\"selected\""
			end
			to_s << ">#@attr_default</option>\n"
		end
		
		@attr_list.each { | item |
			if ( ( @attr_selected == item ) or ( @attr_selected.include?(item) ) ) then
				to_s << "<option selected=\"selected\">"
			else
				to_s << "<option>"
			end
			if @attr_escape == 'true' then
				item = CKMessage.escape( item )
			end
			to_s << "#{item}</option>\n"
		}
		to_s << '</select>'
	end
end


# Creates a checkbox.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: <b>selection</b> and <b>value</b>, or <b>checked</b>
#
# You use this element in two ways. One it the way to use "checked" attribute. 
# The other is the way to use both "selection" and "value" attributes. 
#
# <b>checked</b>::   If neither "value" nor "selection" attribute is nil and
#                    the value of "selection" is equal to that of "value",
#                    the check box is checked.
# <b>value</b>::     When the check box is checked,
#                    the value of "value" attribute is set to the component by
#                    the method specified by "selection" attribute.
# <b>selection</b>:: Object that the user chose from the check box.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKCheckbox < CKElement
	def run
		@attr_name      = oid
		@attr_checked   = fetch( 'checked' )   # true or false
		@attr_value     = fetch( 'value' )     # String
		@attr_selection = fetch( 'selection' ) # String
		@attr_disabled  = fetch( 'disabled' )  # true or false
		if ( ( @attr_value and @attr_selection ) and ( @attr_value == @attr_selection ) ) then
			@attr_checked = true
		end

		if application.attributes_check
			check_undefined_attributes( 'checked', 'value', 'selection', \
				'disabled' )
			check_required_attributes( ['selection','value'], ['checked'] )
		end
	end

	def to_s
		to_s =  "<input type=\"checkbox\" name=\"#@attr_name\""
		to_s << " value=\"#@attr_value\"" if @attr_value
		if @attr_checked  then to_s << " checked" end
		if @attr_disabled then to_s << " disabled" end
		to_s << '>'
	end
end


# Creates a pop-up menu.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>list</b>
#
# <b>escape</b>::    Escapes HTML control characters in the items of the list
#                    if the "escape" is true. The default value is true.
# <b>list</b>::      Array which is iterated through.
# <b>default</b>::   The first item if no item is selected.
# <b>selection</b>:: Item that are chosen from the selection list.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKPopUpButton < CKElement
	def run
		@attr_name     = oid
		@attr_escape   = fetch( 'escape' )      # true or false
		@attr_list     = fetch( 'list' )        # Enumerable
		@attr_default  = fetch( 'default' )     # String
		@attr_selected = fetch( 'selection' )   # String
		@attr_disabled = fetch( 'disabled' )    # true or false

		if application.attributes_check
			check_undefined_attributes( 'escape', 'list', \
				'default', 'selected', 'disabled' )
			check_required_attributes( ['list'] )
		end
	end

	def to_s
		to_s = '<select'
		if @attr_name     then to_s << " name=\"#@attr_name\"" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">\n"
		if @attr_default then
			to_s << "<option value=\"\""
			if @attr_selected == nil then
				to_s << " selected=\"selected\""
			end
			to_s << ">#@attr_default</option>\n"
		end
		
		@attr_list.each { | item |
			to_s << "<option"
			if @attr_selected == item then
				to_s << " selected=\"selected\""
			end
			
			# escape
			if @attr_escape then
				item = CKUtilities.escape_html( item )
			end
			
			to_s << ">#{item}</option>\n"
		}
		to_s << '</select>'
	end
end


# Creates a radio button.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: <b>selection</b> and <b>value</b>, or <b>checked</b>
#
# <b>checked</b>::   If neither "value" nor "selection" attribute is nil and
#                    the value of "selection" is equal to that of "value",
#                    the check box is checked.
# <b>value</b>::     When the check box is checked,
#                    the value of "value" attribute is set to the component by
#                    the method specified by "selection" attribute.
# <b>selection</b>:: Object that the user chose from the check box.
# <b>disabled</b>::  If the value is true, the element appears but is not active.
class CKRadioButton < CKElement
	def run
		@attr_name      = fetch( 'name' )      # String
		@attr_checked   = fetch( 'checked' )   # true or false
		@attr_value     = fetch( 'value' )     # String
		@attr_selection = fetch( 'selection' ) # String
		@attr_disabled  = fetch( 'disabled' )  # true or false
		if ( ( @attr_value and @attr_selection ) and \
			( @attr_value == @attr_selection ) ) then
			@attr_checked = true
		end

		if application.attributes_check
			check_undefined_attributes( 'checked', 'value', \
				'selection', 'disabled', 'name'  )
			check_required_attributes( ['selection','value'], ['checked'] )
		end
	end

	def to_s
		to_s = "<input type=\"radio\" name=\"#@attr_name\" value=\"#@attr_value\""
		if @attr_checked  then to_s << " checked" end
		if @attr_disabled then to_s << " disabled" end
		to_s << ">"
	end
end


# A CKRepeition object repeats its contents. 
#
# == Bindings
# Required attributes: <b>list</b> and <b>item</b>, or <b>count</b>
#
# <b>count</b>:: CKRepeition repeats its contents this number of times 
# <b>list</b>::  Array which is iterated through.
# <b>item</b>::  Current item when the list is iterated through.
# <b>index</b>:: Index of the current item.
class CKRepetition < CKElement
	def run
		@attr_count = fetch( 'count' )        # Integer
		@attr_item  = fetch( 'item', false )
		@attr_list  = fetch( 'list' )         # Enumerable
		@attr_index = fetch( 'index', false )
		if @attr_count == nil then @attr_count = @attr_list.size end

		if application.attributes_check
			check_undefined_attributes( 'count', 'item', 'list', 'index' )
			check_required_attributes( ['list', 'item'], ['count'] )
		end
	end

	def to_s
		to_s  = ''
		index = 0
		parser = CKHTMLParser.new( parent, body )

		@attr_list.each { | item |
			# set item
			CKBinding.bind( parent, @attr_item, item )
			
			# set index
			CKBinding.bind( parent, @attr_index, index )
			
			# parse
			to_s << parser.parse

			if index >= @attr_count then break end
			index += 1
		}
		to_s
	end
end


# Creates a reset button.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>value</b>:: Title of the button.
class CKResetButton < CKElement
	def run
		@attr_value  = fetch( 'value' ) # String
		unless @attr_value then @attr_value = 'Reset' end

		if application.attributes_check
			check_undefined_attributes nil
		end
	end

	def to_s
		"<input type=\"reset\" value=\"#@attr_value\">"
	end
end


# Creates a submit button.
# This element must be used within HTML form.
#
# == Bindings
# Required attributes: none
#
# <b>action</b>::   Method to invoke when the button is clicked.
# <b>value</b>::    Title of the button.
# <b>disabled</b>:: If the value is true, the element appears but is not active.
#                   In addition, it doesn't send the form data 
#                   although the button is clicked.
class CKSubmitButton < CKElement
	def run
		@attr_name     = oid
		@attr_action   = fetch( 'action', false ) # String
		@attr_value    = fetch( 'value' )         # String
		@attr_disabled = fetch( 'disabled' )      # true/false
		if @attr_value == nil then @attr_value = 'Submit' end

		if application.attributes_check
			check_undefined_attributes( 'action', 'value', 'disabled' )
		end
	end

	def to_s
		to_s  = "<input type=\"submit\" name=\"#@attr_name\" value=\"#@attr_value\""
		if @attr_disabled then to_s << " disabled" end
		to_s << ">"
	end
end


# Creates a text area.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>value</b>
#
# <b>value</b>::   Value of the text area. 
# <b>columns</b>:: Column size.
# <b>rows</b>::    Row size.
class CKText < CKElement
	def run
		@attr_name    = oid
		@attr_value   = fetch( 'value' )   # String
		@attr_columns = fetch( 'columns' ) # Integer
		@attr_rows    = fetch( 'rows' )    # Integer

		if application.attributes_check
			check_undefined_attributes( 'value', 'columes', 'rows' )
			check_required_attributes( ['value'] )
		end
	end

	def to_s
		to_s = "<textarea name=\"#@attr_name\""
		if @attr_columns then to_s << " cols=\"#@attr_columns\"" end
		if @attr_rows    then to_s << " rows=\"#@attr_rows\""    end
		to_s << '>'
		if @attr_value   then to_s << @attr_value end
		to_s << '</textarea>'
	end
end


# Creates a text input field.
# This element must be used within CKForm or HTML form.
#
# == Bindings
# Required attribute: <b>value</b>
#
# <b>value</b>:: Value of the text field. If you set an accessor method to
#                this element, the form data is set to a component automatically
#                by the method.
# <b>type</b>::  Type of the text field. "text" is for a normal text input field,
#                "password" is for a password field and 
#                "hidden" is for a hidden field.
# <b>size</b>::  Size of the text field.
class CKTextField < CKElement
	def run
		@attr_name  = oid
		@attr_type  = fetch( 'type' )  # String
		@attr_value = fetch( 'value' ) # String
		@attr_size  = fetch( 'size' )  # Integer
		if @attr_type == nil then @attr_type = 'text' end

		if application.attributes_check
			check_undefined_attributes( 'type', 'value', 'size' )
			check_required_attributes( ['value'] )
		end
	end

	def to_s
		to_s = "<input type=\"#@attr_type\""
		if @attr_name  then to_s << " name=\"#@attr_name\"" end
		if @attr_value then to_s << " value=\"#@attr_value\"" end
		if @attr_size  then to_s << " size=\"#@attr_size\"" end
		to_s << '>'
	end
end


# CKPartsMaker is a Ruby module to deal with a component as parts of web page.
# A component which includes this module isn't displayed even if CGIKit recieves
# requests to show it as the target component ( CKApplication#target ).
#
# A name of parts component is recommended to have "Parts" at
# the end of the name to distinguish it from page component.
#
# == Usage
# Subclass of CKComponent includes CKPartsMaker.
#
#  class AnyParts < CKComponent
#    include CKPartsMaker
#    ...
#  end
#
# == Attributes
# <b>substitute_page</b>:: When CGIKit recieves requests to show component
#                          parts, CGIKit shows the page specified by
#                          this attribute. A main page of an application
#                          is displayed when the value is not defined.
module CKPartsMaker
	attr_accessor :substitute_page
end


