# CKDefinition parses .ckd ( CGIKit Definition ) file.
# CKComponent stores the result of parse
# and uses it for binding.
#
# The format of CKD file is very different from Ruby or HTML
# Here, the format of CKD file is described as pseudo-BNF like Ruby's document.
#
#   STMTS           : (STMS)*
#   STMT            : ELEMENT_NAME `:' ELEMENT_TYPE `{' ATTRS `}'
#   ATTRS           : (ATTRIBUTE `=' VALUE)*
#   ATTRIBUTE       : ruby_method_name
#   VALUE           : METHODS `;'
#                   | LITERAL `;'
#   METHODS         : ruby_method_name(`.'ruby_method_name)*
#   LITERAL         : `true' 
#                   | `false' 
#                   | DIGIT_SEQUENCE 
#                   | STRING 
#   DIGIT_SEQUENCE  : DIGIT+
#   DIGIT           : `0' | `1' | `2' | `3' | `4' | `5' | `6' | `7' | `8' | `9'
#   STRING          : `'' any_character* `''
#                   | `"' any_character* `"'
#   ELEMENT_NAME    : (LETTER | DIGIT | `_')+
#   ELEMENT_TYPE    : ruby_class_name
#   LETTER          : `A' | `B' | `C' | `D' | `E' | `F' | `G' | `H' | `I' | `J' | `K' | `L' | `M' | `N' | `O' | `P' | `Q' | `R' | `S' | `T' | `U' | `V' | `W' | `X' | `Y' | `Z' | `a' | `b' | `c' | `d' | `e' | `f' | `g' | `h' | `i' | `j' | `k' | `l' | `m' | `n' | `o' | `p' | `q' | `r' | `s' | `t' | `u' | `v' | `w' | `x' | `y' | `z'
#
# ruby_method_name means Ruby's Method Name: For example, `foo', `to_s' and `is_a?'.
# Also, ruby_class means Ruby's Class Name: For instance, `CKString' and `CKComponent'.
#
# CKDefinition regards a part which begins from `#' to the end of the line as comment.
# In the next case, `# foo bar comment' is not interpreted by CKDefinition.
#
#   foo : CKString {
#     # foo bar comment
#     value = "foo";
#   }
#
class CKDefinition
	class << self
		def parse_ckd_file( file )
			parse_ckd_string ( _read_ckd_file file )
		end

		def parse_ckd_string( string )
			string      = _delete_comment string
			definitions = {}
			definition  = string.split '}'
			line        = ''

			definition.each { | line |
				myclass, mystruct = line.split '{'

				myclass = myclass.delete " \t\r\n"
				myname, myelement = myclass.split ':'
				definitions[ myname ] = {}
				definitions[ myname ][ 'oid' ]     = myname
				definitions[ myname ][ 'element' ] = myelement

				if mystruct == nil then next end
				mysyntax = mystruct.split ';'

				line2 = ''
				mysyntax.each { | line2 |
					next if line2 =~ /^\#/

					key, value = line2.split '='
					if not ( key and value ) then next end

					key = key.delete " \t\r\n"
					if value =~ /"(.*)"/
						value = "\"#$1\""
					elsif value =~ /'(.*)'"/
						value = "\"#$1\""
					else
						value = value.delete " \t\n"
					end
					definitions[ myname ][ key ] = value
				}
			}

			definitions
		end

		private

		def _read_ckd_file( file )
			f = File.new file
			string = f.readlines.join
			f.close
			string
		end

		def _delete_comment( ckd_string )
			string = ''
			ckd_string.each_line { | line |
				string << line.gsub( /\#(.*)/, '' )
			}
			string
		end
	end
end


# CKHTMLParser parses a template of a component.
# The template is the same as HTML, except one fact that it has
# "<cgikit>" tag. You can use two types of formats as CGIKit tag,
# "<cgikit>...</cgikit>" and "<cgikit />". The difference of the
# formats is whether the tag's body is empty or not. So, if you write
# "<cgikit></cgikit>", it has the same meaning as "<cgikit />".
#
# The CGIKit tag has only one attribute, "name". The "name" attribute
# is a name of CGIKit element. The "name" attribute is composed of "[a-zA-Z0-9_]".
# You can enclose the value of the "name" attribute by double quotation mark.
# And, The value is case-sensitive. For instance, these have the same meaning.
#
#   <cgikit name=foo></cgikit>
#   <cgikit name=foo />
#   <cgikit name="foo"></cgikit>
#   <cgikit name="foo" />
#
# On the other hand, these have different meanings.
#   <cgikit name=Foo />
#   <cgikit name="foo" />
#   <cgikit name=FOO></cgikit>
#
# Comment of HTML is interpreted by CKHTMLParser because it is
# necessary to include something like CSS and Javascript.
class CKHTMLParser
	attr_accessor :application, :html_string, :parent

	def initialize( parent, string )
		@parent      = parent
		@application = @parent.application
		@html_string = string
	end

	def parse
		@_tmp        = @html_string
		@_array      = []
		@_index      = 0
		string = body = tag = ""
		count  = 0
		flag   = false

		_parse_char
		while tag = _next
			if ( _cgikit_tag?( tag ) == true ) or ( flag == true ) \
				and ( _comment_tag?( tag ) == false ) then
				if flag == false then
					body = ''
					name = _name_cgikit_tag( tag )
				end
				flag = true
				body << tag

				if _cgikit_tag?( tag ) then
					if _concluded_tag? tag
						# do nothing
					elsif _open_tag? tag
						count += 1
					else
						count -= 1
					end
				end

				if count == 0 then
					flag    = false
					element = _create_element( name, _body_cgikit_tag( body ) )
					if parsed = element.to_s then string << parsed end
					body = ""
				end
			else
				string << tag
			end
		end
		string
	end
	
	private
	
	def _parse_char
		string = char = ""
		is_tag = false
		is_comment = false

		@_tmp.scan( /./m ) { | char |
			if ( char == "<" ) and ( is_comment == false )
				is_tag = true
				@_array.push( string )
				string = char
			elsif ( char == ">" ) and ( is_tag == true )
				string << ">"
				is_comment = false if string =~ /-->$/
				if is_comment == false
					@_array.push( string )
					is_tag = false
					string = ""
				end
			else
				string << char
				if ( string =~ /^<!--/ ) and ( is_comment == false )
					is_comment = true
				end
			end
			
			prechar = char
		}
		@_array.push( string )
	end
	
	def _next
		@_index += 1
		return @_array[ @_index - 1 ]
	end
	
	def _cgikit_tag?( tag )
		atts = tag.split( " " )
		if atts[0] == nil then return false end
		name = atts[0]
		if ( ( name =~ /<cgikit/i ) or ( name =~ /<\/cgikit/i ) ) then
			return true
		else
			return false
		end
	end
	
	def _concluded_tag?( tag )
		if ( tag =~ /<cgikit/i ) and ( tag =~ /\/>$/ )
			true
		else
			false
		end
	end

	def _open_tag?( tag )
		if tag =~ /<\/cgikit/i
			false
		else
			true
		end
	end
	
	def _name_cgikit_tag( tag )
		atts = tag.split( " ", 2 )
		if atts[0] == nil then return false end
		
		( key, value ) = atts[1].split( "=" )
		value.chop!
		value.tr!("\"\'/\s",'')
		value
	end
	
	def _body_cgikit_tag( body )
		body =~ /<cgikit name=(.*?)>(.*)<\/cgikit>/im
		tagbody = $2
	end

	def _comment_tag?( tag )
		if tag =~ /^<!--/
			true
		else
			false
		end
	end
	
	def _create_element( name, body )
		if parent.definitions[ name ] == nil then return nil end

		element = parent.definitions[ name ][ 'element' ]
		object = CKElement.instance( element, @application, @parent, name, body )
		object.run
		object
	end
end

